from __future__ import annotations

from typing import List, Optional, Sequence, overload

from langchain_core.agents import AgentAction, AgentStep
from langchain_core.callbacks import (
    AsyncCallbackManager,
    CallbackManager,
)
from langchain_core.tools import BaseTool
from langgraph.prebuilt.tool_executor import ToolExecutor  # type: ignore[import-untyped]

from .base_action_executor import BaseActionExecutor
from .meta_tools import MetaTools


class LangchainActionExecutor(BaseActionExecutor):
    def __init__(self, tools: Sequence[BaseTool], meta_tools: Optional[MetaTools] = None):
        self._tool_executor = ToolExecutor(tools)
        self._meta_tool_executor = ToolExecutor(meta_tools.tools) if meta_tools else None
        self._meta_tool_names = meta_tools.tool_names_map if meta_tools else {}

    @property
    def tools(self) -> Sequence[BaseTool]:
        return self._tool_executor.tools

    @property
    def reset_tool_name(self) -> Optional[str]:
        if "reset" in self._meta_tool_names:
            return self._meta_tool_names["reset"]
        return None

    def reset(
        self,
        actions: Optional[List[AgentAction]] = None,
        run_manager: Optional[CallbackManager] = None,
        **kwargs,
    ) -> None:
        """Resets the current state. If actions are passed, will also execute them."""
        if self.reset_tool_name is not None:
            self._execute(
                actions=[
                    AgentAction(
                        tool=self.reset_tool_name,
                        tool_input={},
                        log="Invoking reset tool.",
                    )
                ],
                tool_executor=self._meta_tool_executor,  # type: ignore[reportArgumentType]
                run_manager=run_manager,
            )
            if actions:
                self.execute(actions, run_manager=run_manager)

    @overload
    def execute(
        self,
        actions: List[AgentAction],
        run_manager: Optional[CallbackManager] = None,
        **kwargs,
    ) -> List[AgentStep]: ...

    @overload
    def execute(
        self,
        actions: AgentAction,
        run_manager: Optional[CallbackManager] = None,
        **kwargs,
    ) -> AgentStep: ...

    def execute(
        self,
        actions: List[AgentAction] | AgentAction,
        run_manager: Optional[CallbackManager] = None,
        **kwargs,
    ) -> List[AgentStep] | AgentStep:
        return self._execute(actions, self._tool_executor, run_manager)

    def _execute(
        self,
        actions: List[AgentAction] | AgentAction,
        tool_executor: ToolExecutor,
        run_manager: Optional[CallbackManager] = None,
        **kwargs,
    ) -> List[AgentStep] | AgentStep:
        if isinstance(actions, list):
            steps = []
            for action in actions:
                assert isinstance(action, AgentAction)
                observation = self.execute(action, run_manager=run_manager)
                steps.append(AgentStep(action=action, observation=observation))
            return steps

        assert isinstance(actions, AgentAction)
        observation = tool_executor.invoke(
            actions,
            config={"callbacks": run_manager} if run_manager else {},
        )
        return AgentStep(action=actions, observation=observation)

    async def areset(
        self,
        actions: Optional[List[AgentAction]] = None,
        run_manager: Optional[AsyncCallbackManager] = None,
        **kwargs,
    ) -> None:
        """Resets the current state. If actions are passed, will also execute them."""
        if self.reset_tool_name is not None:
            await self._aexecute(
                actions=[
                    AgentAction(
                        tool=self.reset_tool_name,
                        tool_input={},
                        log="Invoking reset tool.",
                    )
                ],
                tool_executor=self._meta_tool_executor,  # type: ignore[reportArgumentType]
                run_manager=run_manager,
            )
            if actions:
                await self.aexecute(actions, run_manager=run_manager)

    @overload
    async def aexecute(
        self,
        actions: List[AgentAction],
        run_manager: Optional[AsyncCallbackManager] = None,
        **kwargs,
    ) -> List[AgentStep]: ...

    @overload
    async def aexecute(
        self,
        actions: AgentAction,
        run_manager: Optional[AsyncCallbackManager] = None,
        **kwargs,
    ) -> AgentStep: ...

    async def aexecute(
        self,
        actions: List[AgentAction] | AgentAction,
        run_manager: Optional[AsyncCallbackManager] = None,
        **kwargs,
    ) -> List[AgentStep] | AgentStep:
        return await self._aexecute(actions, self._tool_executor, run_manager)

    async def _aexecute(
        self,
        actions: List[AgentAction] | AgentAction,
        tool_executor: ToolExecutor,
        run_manager: Optional[AsyncCallbackManager] = None,
        **kwargs,
    ) -> List[AgentStep] | AgentStep:
        if isinstance(actions, list):
            steps = []
            for action in actions:
                observation = await tool_executor.ainvoke(
                    action,
                    config={"callbacks": run_manager} if run_manager else {},
                )
                steps.append(AgentStep(action=action, observation=observation))
            return steps
        assert isinstance(actions, AgentAction)
        observation = await tool_executor.ainvoke(
            actions,
            config={"callbacks": run_manager} if run_manager else {},
        )
        return AgentStep(action=actions, observation=observation)



import re

from langchain_core.exceptions import OutputParserException
from langchain_core.output_parsers import BaseOutputParser


class SimpleEvaluateOutputParser(BaseOutputParser[float]):
    """Searches for a pattern [[number]] and return the number."""

    def parse(self, text: str) -> float:
        try:
            match = re.search(r"\[\[(.*?)\]\]", text.strip())
            if not match:
                raise ValueError("Pattern [[number]] not found.")
            result = float(match.groups()[0])
            if result < 0.0 or result > 1.0:
                raise ValueError("The given number is out of (0.0, 1.0) range.")
            return result
        except ValueError:
            raise OutputParserException(f"Couldn't convert {text} to float between 0 and 1.")



from .agent_component import AgentComponent
from .base_component import BaseComponent
from .runnable_component import RunnableComponent

__all__ = [
    "BaseComponent",
    "RunnableComponent",
    "AgentComponent",
]



from typing import List, Sequence, Tuple

from langchain.agents.format_scratchpad.openai_tools import (
    format_to_openai_tool_messages,
)
from langchain.agents.output_parsers.openai_tools import OpenAIToolsAgentOutputParser
from langchain_core.agents import AgentAction
from langchain_core.language_models import BaseChatModel
from langchain_core.messages import BaseMessage
from langchain_core.runnables import Runnable
from langchain_core.tools import BaseTool
from langchain_core.utils.function_calling import convert_to_openai_tool

from planning_library.function_calling_parsers.base_parser import (
    AgentInputs,
    BaseFunctionCallingMultiActionParser,
    ProcessedAgentInputs,
)
from planning_library.function_calling_parsers.parser_registry import ParserRegistry


@ParserRegistry.register
class OpenAIToolsParser(BaseFunctionCallingMultiActionParser):
    name: str = "openai-tools"
    output_parser = OpenAIToolsAgentOutputParser()

    @staticmethod
    def _format_intermediate_steps(
        intermediate_steps: List[Tuple[AgentAction, str]],
    ) -> List[BaseMessage]:
        return format_to_openai_tool_messages(intermediate_steps)

    def format_inputs(
        self,
        inputs: AgentInputs,
    ) -> ProcessedAgentInputs:
        return {
            **inputs,  # type: ignore[typeddict-unknown-key]
            "agent_scratchpad": self._format_intermediate_steps(inputs["intermediate_steps"]),
        }

    def prepare_llm(self, llm: BaseChatModel, tools: Sequence[BaseTool]) -> Runnable:
        return llm.bind(tools=[convert_to_openai_tool(tool) for tool in tools])



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List, Optional, Sequence, overload

from langchain_core.agents import AgentAction, AgentStep
from langchain_core.callbacks import AsyncCallbackManager, CallbackManager
from langchain_core.tools import BaseTool


class BaseActionExecutor(ABC):
    @property
    @abstractmethod
    def tools(self) -> Sequence[BaseTool]: ...

    @abstractmethod
    def reset(
        self,
        actions: Optional[List[AgentAction]] = None,
        run_manager: Optional[CallbackManager] = None,
        **kwargs,
    ) -> None:
        """Resets the current state. If actions are passed, will also execute them."""
        ...

    @abstractmethod
    async def areset(
        self,
        actions: Optional[List[AgentAction]] = None,
        run_manager: Optional[AsyncCallbackManager] = None,
        **kwargs,
    ) -> None:
        """Resets the current state. If actions are passed, will also execute them."""
        ...

    @overload
    @abstractmethod
    def execute(
        self,
        actions: List[AgentAction],
        run_manager: Optional[CallbackManager] = None,
        **kwargs,
    ) -> List[AgentStep]: ...

    @overload
    @abstractmethod
    def execute(
        self,
        actions: AgentAction,
        run_manager: Optional[CallbackManager] = None,
        **kwargs,
    ) -> AgentStep: ...

    @abstractmethod
    def execute(
        self,
        actions: List[AgentAction] | AgentAction,
        run_manager: Optional[CallbackManager] = None,
        **kwargs,
    ) -> List[AgentStep] | AgentStep:
        """Performs actions.

        Args:
            actions: Currently proposed actions. Can be: multi-action, single action.

        Returns:
              * List[AgentStep] - for multi-action thoughts (List[AgentAction])
              * AgentStep - for single-action thoughts (AgentAction)
        """
        ...

    @overload
    @abstractmethod
    async def aexecute(
        self,
        actions: List[AgentAction],
        run_manager: Optional[AsyncCallbackManager] = None,
        **kwargs,
    ) -> List[AgentStep]: ...

    @overload
    @abstractmethod
    async def aexecute(
        self,
        actions: AgentAction,
        run_manager: Optional[AsyncCallbackManager] = None,
        **kwargs,
    ) -> AgentStep: ...

    @abstractmethod
    async def aexecute(
        self,
        actions: List[AgentAction] | AgentAction,
        run_manager: Optional[AsyncCallbackManager] = None,
        **kwargs,
    ) -> List[AgentStep] | AgentStep:
        """Performs actions asynchronously.

        Args:
            actions: Currently proposed actions. Can be: multi-action, single action.

        Returns:
              * List[AgentStep] - for multi-action thoughts (List[AgentAction])
              * AgentStep - for single-action thoughts (AgentAction)
        """
        ...



from .planner_output_parser import SimplePlannerOutputParser
from .typing_utils import ADaPTPlan, ADaPTPlannerInput, ADaPTPlannerOutput

__all__ = [
    "ADaPTPlan",
    "ADaPTPlannerOutput",
    "ADaPTPlannerInput",
    "SimplePlannerOutputParser",
]



from typing import Any, Dict, Optional, Sequence, Tuple

from langchain.agents.tools import InvalidTool
from langchain_core.agents import AgentAction, AgentStep
from langchain_core.callbacks import (
    AsyncCallbackManagerForChainRun,
    CallbackManagerForChainRun,
)
from langchain_core.tools import BaseTool
from langchain_core.utils.input import get_color_mapping


def get_tools_maps(
    tools: Sequence[BaseTool],
) -> Tuple[Dict[str, BaseTool], Dict[str, str]]:
    name_to_tool_map = {tool.name: tool for tool in tools}
    color_mapping = get_color_mapping([tool.name for tool in tools], excluded_colors=["green"])
    return name_to_tool_map, color_mapping


def perform_agent_action(
    agent_action: AgentAction,
    name_to_tool_map: Dict[str, BaseTool],
    color_mapping: Dict[str, str],
    verbose: bool = True,
    tool_run_kwargs: Optional[Dict[str, Any]] = None,
    run_manager: Optional[CallbackManagerForChainRun] = None,
) -> AgentStep:
    """
    Copied from langchain.agents.agent.AgentExecutor._perform_agent_action with slight modifications.
    """

    tool_run_kwargs = {} if tool_run_kwargs is None else tool_run_kwargs

    if run_manager:
        run_manager.on_agent_action(agent_action, color="green")
    # Otherwise we lookup the tool
    if agent_action.tool in name_to_tool_map:
        tool = name_to_tool_map[agent_action.tool]
        return_direct = tool.return_direct
        color = color_mapping[agent_action.tool]
        if return_direct:
            tool_run_kwargs["llm_prefix"] = ""

        # We then call the tool on the tool input to get an observation
        observation = tool.run(
            agent_action.tool_input,
            verbose=verbose,
            color=color,
            callbacks=run_manager.get_child() if run_manager else None,
            **tool_run_kwargs,
        )
    else:
        observation = InvalidTool().run(
            {
                "requested_tool_name": agent_action.tool,
                "available_tool_names": list(name_to_tool_map.keys()),
            },
            verbose=verbose,
            color=None,
            callbacks=run_manager.get_child() if run_manager else None,
            **tool_run_kwargs,
        )
    return AgentStep(action=agent_action, observation=observation)


async def aperform_agent_action(
    name_to_tool_map: Dict[str, BaseTool],
    color_mapping: Dict[str, str],
    agent_action: AgentAction,
    verbose: bool = True,
    tool_run_kwargs: Optional[Dict[str, Any]] = None,
    run_manager: Optional[AsyncCallbackManagerForChainRun] = None,
) -> AgentStep:
    """
    Copied from langchain.agents.agent.AgentExecutor._aperform_agent_action with slight modifications.
    """
    tool_run_kwargs = {} if tool_run_kwargs is None else tool_run_kwargs

    if run_manager:
        await run_manager.on_agent_action(agent_action, verbose=verbose, color="green")
    # Otherwise we lookup the tool
    if agent_action.tool in name_to_tool_map:
        tool = name_to_tool_map[agent_action.tool]
        return_direct = tool.return_direct
        color = color_mapping[agent_action.tool]
        if return_direct:
            tool_run_kwargs["llm_prefix"] = ""
        # We then call the tool on the tool input to get an observation
        observation = await tool.arun(
            agent_action.tool_input,
            verbose=verbose,
            color=color,
            callbacks=run_manager.get_child() if run_manager else None,
            **tool_run_kwargs,
        )
    else:
        observation = await InvalidTool().arun(
            {
                "requested_tool_name": agent_action.tool,
                "available_tool_names": list(name_to_tool_map.keys()),
            },
            verbose=verbose,
            color=None,
            callbacks=run_manager.get_child() if run_manager else None,
            **tool_run_kwargs,
        )
    return AgentStep(action=agent_action, observation=observation)



import asyncio
from abc import ABC, abstractmethod
from typing import (
    Any,
    AsyncIterator,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
)

from langchain.chains.base import Chain
from langchain_core.agents import AgentAction, AgentFinish
from langchain_core.callbacks import (
    AsyncCallbackManagerForChainRun,
    CallbackManagerForChainRun,
)


class BaseCustomStrategy(Chain, ABC):
    return_intermediate_steps: bool = False
    return_finish_log: bool = False
    max_iterations: int = 15
    verbose: bool = True

    @property
    def input_keys(self) -> List[str]:
        """Keys expected to be in the chain input."""
        return []

    @property
    def output_keys(self) -> List[str]:
        """Keys expected to be in the chain output."""
        return []

    @abstractmethod
    def _run_strategy(
        self,
        inputs: Dict[str, str],
        run_manager: Optional[CallbackManagerForChainRun] = None,
    ) -> Iterator[Tuple[AgentFinish, List[Tuple[AgentAction, str]]]]: ...

    async def _arun_strategy(
        self,
        inputs: Dict[str, str],
        run_manager: Optional[AsyncCallbackManagerForChainRun] = None,
    ) -> AsyncIterator[Tuple[AgentFinish, List[Tuple[AgentAction, str]]]]:
        loop = asyncio.get_event_loop()

        sync_run_manager = run_manager.get_sync() if run_manager is not None else None
        result = await loop.run_in_executor(None, self._run_strategy, inputs, sync_run_manager)
        for item in result:
            yield item

    def _return(
        self,
        output: AgentFinish,
        intermediate_steps: list,
        run_manager: Optional[CallbackManagerForChainRun] = None,
    ) -> Dict[str, Any]:
        if run_manager:
            run_manager.on_agent_finish(output, color="green", verbose=self.verbose)
        final_output = output.return_values
        if self.return_intermediate_steps:
            final_output["intermediate_steps"] = intermediate_steps
        if self.return_finish_log:
            final_output["finish_log"] = output.log
        return final_output

    async def _areturn(
        self,
        output: AgentFinish,
        intermediate_steps: list,
        run_manager: Optional[AsyncCallbackManagerForChainRun] = None,
    ) -> Dict[str, Any]:
        if run_manager:
            await run_manager.on_agent_finish(output, color="green", verbose=self.verbose)
        final_output = output.return_values
        if self.return_intermediate_steps:
            final_output["intermediate_steps"] = intermediate_steps
        if self.return_finish_log:
            final_output["finish_log"] = output.log
        return final_output

    def _call(
        self,
        inputs: Dict[str, Any],
        run_manager: Optional[CallbackManagerForChainRun] = None,
    ) -> Dict[str, Any]:
        """Run text through and get agent response."""

        outputs = [
            self._return(output, intermediate_steps, run_manager=run_manager)
            for output, intermediate_steps in self._run_strategy(
                inputs=inputs,
                run_manager=run_manager,
            )
        ]

        return {key: [output[key] for output in outputs] for key in outputs[0]}

    async def _acall(
        self,
        inputs: Dict[str, Any],
        run_manager: Optional[AsyncCallbackManagerForChainRun] = None,
    ) -> Dict[str, Any]:
        """Run text through and get agent response."""

        _outputs = self._arun_strategy(
            inputs=inputs,
            run_manager=run_manager,
        )

        outputs = []
        async for _output, _intermediate_steps in _outputs:
            output = await self._areturn(_output, _intermediate_steps, run_manager=run_manager)
            outputs.append(output)

        return {key: [output[key] for output in outputs] for key in outputs[0]}


# TODO: what should the interface be?
class BaseLangGraphStrategy(ABC): ...



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Sequence, Tuple

from langchain.agents.agent import AgentOutputParser, MultiActionAgentOutputParser
from langchain_core.agents import AgentAction
from langchain_core.language_models import BaseChatModel
from langchain_core.messages import BaseMessage
from langchain_core.runnables import Runnable
from langchain_core.tools import BaseTool
from typing_extensions import TypedDict


class AgentInputs(TypedDict):
    inputs: Dict[str, Any]
    intermediate_steps: List[Tuple[AgentAction, str]]


class ProcessedAgentInputs(TypedDict):
    inputs: Dict[str, Any]
    agent_scratchpad: List[BaseMessage]


class BaseFunctionCallingParser(ABC):
    name: str

    @abstractmethod
    def prepare_llm(self, llm: BaseChatModel, tools: Sequence[BaseTool]) -> Runnable: ...

    @abstractmethod
    def format_inputs(self, inputs: AgentInputs) -> ProcessedAgentInputs: ...


class BaseFunctionCallingSingleActionParser(BaseFunctionCallingParser, ABC):
    output_parser: AgentOutputParser


class BaseFunctionCallingMultiActionParser(BaseFunctionCallingParser, ABC):
    output_parser: MultiActionAgentOutputParser



from __future__ import annotations

from textwrap import dedent
from typing import Any, Dict, List, Optional, Sequence, Tuple, Type

from langchain_core.agents import AgentAction, AgentFinish
from langchain_core.language_models import BaseChatModel
from langchain_core.messages import BaseMessage
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from typing_extensions import TypedDict

from planning_library.components import RunnableComponent
from planning_library.function_calling_parsers import (
    BaseFunctionCallingMultiActionParser,
    BaseFunctionCallingSingleActionParser,
    ParserRegistry,
)


class ReflexionSelfReflectionInput(TypedDict):
    inputs: Dict[str, Any]
    intermediate_steps: List[Tuple[AgentAction, str]]
    agent_outcome: AgentFinish


class PreprocessedReflexionSelfReflectionInput(TypedDict):
    intermediate_steps: List[BaseMessage]
    agent_outcome: str


class ReflexionSelfReflection(RunnableComponent[ReflexionSelfReflectionInput, Sequence[BaseMessage]]):
    name = "Self-Reflection"

    required_prompt_input_vars = set(ReflexionSelfReflectionInput.__annotations__) - {"inputs"}

    @classmethod
    def _create_default_prompt(cls, system_message: Optional[str], user_message: str, **kwRGS) -> ChatPromptTemplate:
        if system_message is None:
            system_message = "You are an advanced reasoning agent that can self-reflect on their shortcomings when solving reasoning tasks."

        return ChatPromptTemplate.from_messages(
            [
                ("system", system_message),
                (
                    "human",
                    user_message,
                ),
                MessagesPlaceholder("intermediate_steps"),
                ("human", "Final Outcome: {agent_outcome}"),
                (
                    "human",
                    dedent("""
                     In this trial, you were unsuccessful. 
                     In a few sentences, diagnose a possible reason for failure and devise a new, concise, high level plan that aims to mitigate the same shortcomings. 
                     Use complete sentences.
                     """),
                ),
            ]
        )

    @classmethod
    def create(
        cls: Type["ReflexionSelfReflection"],
        llm: BaseChatModel,
        prompt: Optional[ChatPromptTemplate] = None,
        user_message: Optional[str] = None,
        system_message: Optional[str] = None,
        parser: Optional[BaseFunctionCallingSingleActionParser | BaseFunctionCallingMultiActionParser] = None,
        parser_name: Optional[str] = None,
    ) -> "ReflexionSelfReflection":
        def _preprocess_input(
            inputs: ReflexionSelfReflectionInput,
        ) -> Dict:
            nonlocal parser, parser_name
            if parser is None:
                assert parser_name is not None
                parser = ParserRegistry.get_parser(parser_name)

            preprocessed_inputs = parser.format_inputs(inputs)
            return {
                **preprocessed_inputs["inputs"],
                "agent_outcome": preprocessed_inputs["agent_outcome"].return_values["output"],  # type: ignore[typeddict-item]
                "intermediate_steps": preprocessed_inputs["agent_scratchpad"],
            }

        prompt = cls._process_prompt(prompt=prompt, user_message=user_message, system_message=system_message)

        # TODO: figure out typing here
        self_reflection: ReflexionSelfReflection = cls.create_from_steps(  # type: ignore[assignment]
            prompt=prompt, llm=llm
        )
        self_reflection.add_input_preprocessing(_preprocess_input)
        return self_reflection



from abc import ABC, abstractmethod
from typing import Awaitable, Callable, Dict, Generic, Mapping, Optional, Set, TypeVar

from langchain_core.callbacks import AsyncCallbackManager, CallbackManager
from langchain_core.prompts import ChatPromptTemplate

InputType = TypeVar("InputType", bound=Mapping)
OutputType = TypeVar("OutputType")


class BaseComponent(Generic[InputType, OutputType], ABC):
    name: Optional[str] = None
    required_prompt_input_vars: Set[str] = set()

    @classmethod
    def _create_default_prompt(cls, system_message: Optional[str], user_message: str, **kwargs) -> ChatPromptTemplate:
        raise NotImplementedError(
            f"Default prompt is not supported for {cls.__name__}. Please, provide `prompt` instead of `user_message`."
        )

    @classmethod
    def _process_prompt(
        cls,
        prompt: Optional[ChatPromptTemplate] = None,
        system_message: Optional[str] = None,
        user_message: Optional[str] = None,
        **kwargs,
    ) -> ChatPromptTemplate:
        if prompt is None:
            if user_message is None:
                raise ValueError("Either `prompt` or `user_message` are required to create an agent.")
            prompt = cls._create_default_prompt(system_message=system_message, user_message=user_message, **kwargs)

        missing_vars = cls.required_prompt_input_vars.difference(prompt.input_variables)
        if missing_vars:
            raise ValueError(f"Prompt for {cls.__name__} missing required variables: {missing_vars}")

        return prompt

    def add_input_preprocessing(
        self,
        preprocess: Callable[[InputType], Dict],
        apreprocess: Optional[Callable[[InputType], Awaitable[Dict]]] = None,
    ) -> None: ...

    def add_output_preprocessing(
        self,
        preprocess: Callable[[OutputType], OutputType],
        apreprocess: Optional[Callable[[OutputType], Awaitable[OutputType]]] = None,
    ) -> None: ...

    @abstractmethod
    def invoke(self, inputs: InputType, run_manager: Optional[CallbackManager] = None, **kwargs) -> OutputType: ...

    @abstractmethod
    async def ainvoke(
        self,
        inputs: InputType,
        run_manager: Optional[AsyncCallbackManager] = None,
        **kwargs,
    ) -> OutputType: ...



from .base_parser import (
    BaseFunctionCallingMultiActionParser,
    BaseFunctionCallingSingleActionParser,
)
from .openai_functions_parser import OpenAIFunctionsParser
from .openai_tools_parser import OpenAIToolsParser
from .parser_registry import ParserRegistry

__all__ = [
    "BaseFunctionCallingMultiActionParser",
    "BaseFunctionCallingSingleActionParser",
    "OpenAIToolsParser",
    "OpenAIFunctionsParser",
    "ParserRegistry",
]



from typing import Awaitable, Callable, Dict, Optional

from langchain_core.callbacks import AsyncCallbackManager, CallbackManager
from langchain_core.language_models import BaseChatModel
from langchain_core.output_parsers import BaseOutputParser
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.runnables import Runnable, RunnableLambda

from .base_component import BaseComponent, InputType, OutputType


class RunnableComponent(BaseComponent[InputType, OutputType]):
    def __init__(self, runnable: Runnable[InputType, OutputType]):
        self.runnable = runnable

    @classmethod
    def create_from_steps(
        cls,
        llm: BaseChatModel,
        output_parser: Optional[BaseOutputParser] = None,
        prompt: Optional[ChatPromptTemplate] = None,
        user_message: Optional[str] = None,
        system_message: Optional[str] = None,
    ) -> "RunnableComponent":
        prompt = cls._process_prompt(prompt=prompt, user_message=user_message, system_message=system_message)
        runnable = prompt | llm
        if output_parser is not None:
            runnable = runnable | output_parser
        return RunnableComponent(runnable)

    def add_input_preprocessing(
        self,
        preprocess: Callable[[InputType], Dict],
        apreprocess: Optional[Callable[[InputType], Awaitable[Dict]]] = None,
    ) -> None:
        self.runnable = RunnableLambda(preprocess, afunc=apreprocess) | self.runnable

    def add_output_preprocessing(
        self,
        preprocess: Callable[[OutputType], OutputType],
        apreprocess: Optional[Callable[[OutputType], Awaitable[OutputType]]] = None,
    ) -> None:
        self.runnable = self.runnable | RunnableLambda(preprocess, afunc=apreprocess)

    def invoke(self, inputs: InputType, run_manager: Optional[CallbackManager] = None, **kwargs) -> OutputType:
        config = kwargs
        if "callbacks" not in config and run_manager:
            config["callbacks"] = run_manager

        if "run_name" not in config and self.name:
            config["run_name"] = self.name

        outputs = self.runnable.invoke(
            inputs,
            config=config,  # type: ignore[arg-type]
        )
        return outputs

    async def ainvoke(
        self,
        inputs: InputType,
        run_manager: Optional[AsyncCallbackManager] = None,
        **kwargs,
    ) -> OutputType:
        config = kwargs
        if "callbacks" not in config and run_manager:
            config["callbacks"] = run_manager

        if "run_name" not in config and self.name:
            config["run_name"] = self.name

        outputs = await self.runnable.ainvoke(
            inputs,
            config=config,  # type: ignore[arg-type]
        )
        return outputs



from __future__ import annotations

from typing import Awaitable, Callable, Dict, List, Optional, Sequence, Union

from langchain.agents.agent import BaseMultiActionAgent, BaseSingleActionAgent, RunnableAgent, RunnableMultiActionAgent
from langchain_core.agents import AgentAction, AgentFinish
from langchain_core.callbacks import AsyncCallbackManager, CallbackManager
from langchain_core.language_models import BaseChatModel
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.runnables import Runnable, RunnableLambda
from langchain_core.tools import BaseTool

from planning_library.function_calling_parsers import (
    BaseFunctionCallingMultiActionParser,
    BaseFunctionCallingSingleActionParser,
    ParserRegistry,
)
from planning_library.utils import (
    convert_runnable_to_agent,
)

from .base_component import BaseComponent, InputType


class AgentFactory:
    @staticmethod
    def create_agent(
        llm: BaseChatModel,
        tools: Sequence[BaseTool],
        prompt: ChatPromptTemplate,
        parser: Optional[
            Union[
                BaseFunctionCallingSingleActionParser,
                BaseFunctionCallingMultiActionParser,
            ]
        ] = None,
        parser_name: Optional[str] = None,
    ) -> Union[RunnableAgent, RunnableMultiActionAgent]:
        if parser is None:
            if parser_name is None:
                raise ValueError("Either parser or parser_name should be provided to instantiate an agent.")
            parser = ParserRegistry.get_parser(parser_name)

        llm_with_tools = parser.prepare_llm(llm=llm, tools=tools)

        runnable: Runnable = RunnableLambda(parser.format_inputs) | prompt | llm_with_tools | parser.output_parser
        agent = convert_runnable_to_agent(runnable)
        return agent


class AgentComponent(BaseComponent[InputType, Union[List[AgentAction], AgentAction, AgentFinish]]):
    def __init__(
        self,
        agent: BaseSingleActionAgent | BaseMultiActionAgent,
        flatten_inputs: bool = True,
    ):
        self.agent = agent
        if flatten_inputs:
            self.add_input_preprocessing(
                lambda x: {
                    **x["inputs"],
                    **{key: x[key] for key in x if key != "inputs"},
                }
            )

    def add_input_preprocessing(
        self,
        preprocess: Callable[[InputType], Dict],
        apreprocess: Optional[Callable[[InputType], Awaitable[Dict]]] = None,
    ) -> None:
        if hasattr(self.agent, "runnable"):
            self.agent.runnable = RunnableLambda(preprocess, afunc=apreprocess) | self.agent.runnable  # type: ignore[reportAttributeAccessIssue]

    def add_output_preprocessing(
        self,
        preprocess: Callable[
            [Union[List[AgentAction], AgentAction, AgentFinish]],
            Union[List[AgentAction], AgentAction, AgentFinish],
        ],
        apreprocess: Optional[
            Callable[
                [Union[List[AgentAction], AgentAction, AgentFinish]],
                Awaitable[Union[List[AgentAction], AgentAction, AgentFinish]],
            ]
        ] = None,
    ) -> None:
        if hasattr(self.agent, "runnable"):
            self.agent.runnable = self.agent.runnable | RunnableLambda(preprocess, afunc=apreprocess)  # type: ignore[reportAttributeAccessIssue]

    def invoke(
        self, inputs: InputType, run_manager: Optional[CallbackManager] = None, **kwargs
    ) -> Union[List[AgentAction], AgentAction, AgentFinish]:
        # TODO: no way to pass name to plan?
        # TODO: intermediate_steps?
        return self.agent.plan(**inputs, callbacks=run_manager)  # type: ignore[reportCallIssue]

    async def ainvoke(
        self,
        inputs: InputType,
        run_manager: Optional[AsyncCallbackManager] = None,
        **kwargs,
    ) -> Union[List[AgentAction], AgentAction, AgentFinish]:
        # TODO: no way to pass name to plan?
        outputs = await self.agent.aplan(**inputs, callbacks=run_manager)  # type: ignore[reportCallIssue]
        return outputs

    @classmethod
    def create(
        cls,
        llm: BaseChatModel,
        tools: Sequence[BaseTool],
        prompt: Optional[ChatPromptTemplate] = None,
        user_message: Optional[str] = None,
        system_message: Optional[str] = None,
        parser: Optional[
            Union[
                BaseFunctionCallingSingleActionParser,
                BaseFunctionCallingMultiActionParser,
            ]
        ] = None,
        parser_name: Optional[str] = None,
    ) -> "AgentComponent[InputType]":
        prompt = cls._process_prompt(prompt=prompt, user_message=user_message, system_message=system_message)

        return cls(
            agent=AgentFactory.create_agent(
                llm=llm,
                tools=tools,
                prompt=prompt,
                parser=parser,
                parser_name=parser_name,
            )
        )



from .actions_utils import aperform_agent_action, get_tools_maps, perform_agent_action
from .convert_runnable_to_agent import convert_runnable_to_agent
from .format_agent_outputs import format_thought, format_thoughts

__all__ = [
    "convert_runnable_to_agent",
    "perform_agent_action",
    "aperform_agent_action",
    "get_tools_maps",
    "format_thought",
    "format_thoughts",
]



from __future__ import annotations

from typing import Any, AsyncIterator, Dict, Iterator, List, Optional, Tuple

from langchain_core.agents import AgentAction, AgentFinish
from langchain_core.callbacks import (
    AsyncCallbackManagerForChainRun,
    CallbackManagerForChainRun,
)

from planning_library.strategies.adapt.components import ADaPTExecutor, ADaPTPlanner

from ..base_strategy import BaseCustomStrategy


class ADaPTStrategy(BaseCustomStrategy):
    """ADaPT strategy.

    Based on "ADaPT: As-Needed Decomposition and Planning with Language Models" by Prasad et al.
    """

    executor: ADaPTExecutor
    planner: ADaPTPlanner
    max_depth: int

    @property
    def input_keys(self) -> List[str]:
        # TODO: define properly
        return []

    @property
    def output_keys(self) -> List[str]:
        # TODO: define properly
        return []

    def _adapt_step(
        self,
        inputs: Dict[str, Any],
        depth: int,
        intermediate_steps: List[Tuple[AgentAction, str]],
        run_manager: Optional[CallbackManagerForChainRun] = None,
    ) -> Tuple[bool, AgentFinish, List[Tuple[AgentAction, str]]]:
        """Performs an iteration of ADaPT strategy.

        Args:
            current_task: The input for the current iteration. It can either be the original input or a subtask of a plan generated on a previous step.
            intermediate_steps: A list of actions taken before the current iteration.
            run_manager: Callback for the current run.
        """
        # 1: if we're too deep in task decomposition, finish early
        if depth > self.max_depth:
            return (
                False,
                AgentFinish(return_values={}, log="Maximum decomposition depth reached."),
                intermediate_steps,
            )

        # 2: run task through executor
        executor_output = self.executor.invoke(
            inputs,  # type: ignore
            run_manager=run_manager.get_child(tag=f"executor:depth_{depth}") if run_manager else None,
        )

        is_completed, cur_agent_outcome, cur_intermediate_steps = (
            executor_output["is_completed"],
            executor_output["agent_outcome"],
            executor_output["intermediate_steps"],
        )

        # 3.1: if executor estimated successful completion of a task, wrap up
        if is_completed:
            intermediate_steps.extend(cur_intermediate_steps)
            return True, cur_agent_outcome, intermediate_steps
        else:
            # 3.2: otherwise:
            self.executor.reset(
                actions=[a[0] for a in intermediate_steps],
                run_manager=run_manager.get_child(tag="clean_env") if run_manager else None,
            )

            # call a planner to further decompose a current task
            plan = self.planner.invoke(
                dict(
                    inputs=inputs,  # type: ignore[reportArgumentType]
                    executor_agent_outcome=cur_agent_outcome,
                    executor_intermediate_steps=cur_intermediate_steps,
                ),
                run_manager=run_manager.get_child(tag=f"executor:depth_{depth}") if run_manager else None,
            )
            # when AND logic is given, execute tasks sequentially
            if plan["aggregation_mode"] == "and":
                for task_inputs in plan["subtasks"]:
                    cur_is_completed, cur_agent_outcome, cur_intermediate_steps = self._adapt_step(
                        inputs={"inputs": task_inputs},
                        depth=depth + 1,
                        run_manager=run_manager,
                        intermediate_steps=intermediate_steps,
                    )

                    if not cur_is_completed:
                        agent_outcome = AgentFinish(
                            return_values=cur_agent_outcome.return_values,
                            log=f"Couldn't solve the task. Last log: {cur_agent_outcome.log}",
                        )
                        return False, agent_outcome, intermediate_steps
                    else:
                        intermediate_steps.extend(cur_intermediate_steps)

                agent_outcome = AgentFinish(return_values={}, log="Task solved successfully!")
                return True, agent_outcome, intermediate_steps
            elif plan["aggregation_mode"] == "or":
                for task_inputs in plan["subtasks"]:
                    (
                        cur_is_completed,
                        cur_agent_outcome,
                        cur_intermediate_steps,
                    ) = self._adapt_step(
                        inputs={"inputs": task_inputs},  # TODO: hard-coded inputs key is ugly
                        depth=depth + 1,
                        run_manager=run_manager,
                        intermediate_steps=intermediate_steps,
                    )

                    if cur_is_completed:
                        agent_outcome = AgentFinish(return_values={}, log="Task solved successfully!")
                        return True, agent_outcome, intermediate_steps
                    else:
                        intermediate_steps.extend(cur_intermediate_steps)

                agent_outcome = AgentFinish(
                    return_values=cur_agent_outcome.return_values,
                    log=f"Couldn't solve the task. Last log: {cur_agent_outcome.log}",
                )
                return False, agent_outcome, intermediate_steps

            raise NotImplementedError("Currently, only `and` and `or` aggregation logic is supported.")

    def _run_strategy(
        self,
        inputs: Dict[str, str],
        run_manager: Optional[CallbackManagerForChainRun] = None,
    ) -> Iterator[Tuple[AgentFinish, List[Tuple[AgentAction, str]]]]:
        _, agent_outcome, intermediate_steps = self._adapt_step(
            inputs=inputs,
            depth=0,
            run_manager=run_manager,
            intermediate_steps=[],
        )
        yield agent_outcome, intermediate_steps

    async def _adapt_astep(
        self,
        inputs: Dict[str, Any],
        depth: int,
        intermediate_steps: List[Tuple[AgentAction, str]],
        run_manager: Optional[AsyncCallbackManagerForChainRun] = None,
    ) -> Tuple[bool, AgentFinish, List[Tuple[AgentAction, str]]]:
        """Performs an iteration of ADaPT strategy asynchronously.

        Args:
            inputs: The input for the current iteration. It can either be the original input or a subtask of a plan generated on a previous step.
            depth: Current decomposition step.
            intermediate_steps: A list of actions taken before the current iteration.
            run_manager: Callback for the current run.
        """
        # 1: if we're too deep in task decomposition, finish early
        if depth > self.max_depth:
            return (
                False,
                AgentFinish(return_values={}, log="Maximum decomposition depth reached."),
                intermediate_steps,
            )

        # 2: run task through executor
        executor_output = await self.executor.ainvoke(
            dict(
                inputs=inputs,  # type: ignore
            ),
            run_manager=run_manager.get_child(tag=f"executor:depth_{depth}") if run_manager else None,
        )
        is_completed, cur_agent_outcome, cur_intermediate_steps = (
            executor_output["is_completed"],
            executor_output["agent_outcome"],
            executor_output["intermediate_steps"],
        )

        # 3.1: if executor estimated successful completion of a task, wrap up
        if is_completed:
            intermediate_steps.extend(cur_intermediate_steps)
            return True, cur_agent_outcome, intermediate_steps
        else:
            # 3.2: otherwise:
            await self.executor.areset(
                actions=[a[0] for a in intermediate_steps],
                run_manager=run_manager.get_child(tag="clean_env") if run_manager else None,
            )

            plan = await self.planner.ainvoke(
                dict(
                    inputs=inputs,  # type: ignore[reportArgumentType]
                    executor_agent_outcome=cur_agent_outcome,
                    executor_intermediate_steps=cur_intermediate_steps,
                ),
                run_manager=run_manager.get_child(tag=f"executor:depth_{depth}") if run_manager else None,
            )
            # when AND logic is given, execute tasks sequentially
            if plan["aggregation_mode"] == "and":
                for task_inputs in plan["subtasks"]:
                    (
                        cur_is_completed,
                        cur_agent_outcome,
                        cur_intermediate_steps,
                    ) = await self._adapt_astep(
                        inputs={"inputs": task_inputs},
                        depth=depth + 1,
                        run_manager=run_manager,
                        intermediate_steps=intermediate_steps,
                    )

                    if not cur_is_completed:
                        agent_outcome = AgentFinish(
                            return_values=cur_agent_outcome.return_values,
                            log=f"Couldn't solve the task. Last log: {cur_agent_outcome.log}",
                        )
                        return False, agent_outcome, intermediate_steps
                    else:
                        intermediate_steps.extend(cur_intermediate_steps)

                agent_outcome = AgentFinish(return_values={}, log="Task solved successfully!")
                return True, agent_outcome, intermediate_steps
            elif plan["aggregation_mode"] == "or":
                for task_inputs in plan["subtasks"]:
                    (
                        cur_is_completed,
                        cur_agent_outcome,
                        cur_intermediate_steps,
                    ) = await self._adapt_astep(
                        inputs={"inputs": task_inputs},
                        depth=depth + 1,
                        run_manager=run_manager,
                        intermediate_steps=intermediate_steps,
                    )

                    if cur_is_completed:
                        agent_outcome = AgentFinish(return_values={}, log="Task solved successfully!")
                        return True, agent_outcome, intermediate_steps
                    else:
                        intermediate_steps.extend(cur_intermediate_steps)

                agent_outcome = AgentFinish(
                    return_values=cur_agent_outcome.return_values,
                    log=f"Couldn't solve the task. Last log: {cur_agent_outcome.log}",
                )
                return False, agent_outcome, intermediate_steps

            raise NotImplementedError("Currently, only `and` and `or` aggregation logic is supported.")

    async def _arun_strategy(
        self,
        inputs: Dict[str, str],
        run_manager: Optional[AsyncCallbackManagerForChainRun] = None,
    ) -> AsyncIterator[Tuple[AgentFinish, List[Tuple[AgentAction, str]]]]:
        _, agent_outcome, intermediate_steps = await self._adapt_astep(
            inputs=inputs,
            depth=0,
            run_manager=run_manager,
            intermediate_steps=[],
        )
        yield agent_outcome, intermediate_steps



from __future__ import annotations

from textwrap import dedent
from typing import Any, Dict, Optional, Union

from langchain_core.callbacks import AsyncCallbackManager, CallbackManager
from langchain_core.language_models import BaseChatModel
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain_core.runnables import Runnable

from planning_library.components import BaseComponent, RunnableComponent
from planning_library.function_calling_parsers import (
    BaseFunctionCallingMultiActionParser,
    BaseFunctionCallingSingleActionParser,
    ParserRegistry,
)
from planning_library.strategies.adapt.utils import ADaPTPlannerInput, ADaPTPlannerOutput, SimplePlannerOutputParser
from planning_library.utils import format_thought


class ADaPTPlanner(BaseComponent[ADaPTPlannerInput, ADaPTPlannerOutput]):
    name = "Planner"

    def __init__(
        self,
        runnable: Runnable[ADaPTPlannerInput, ADaPTPlannerOutput]
        | RunnableComponent[ADaPTPlannerInput, ADaPTPlannerOutput],
    ):
        if not isinstance(runnable, RunnableComponent):
            runnable = RunnableComponent(runnable)
        self.runnable = runnable

    @classmethod
    def _create_default_prompt(
        cls,
        system_message: Optional[str],
        user_message: str,
        **kwargs,
    ) -> ChatPromptTemplate:
        if system_message is None:
            system_message = (
                "You are an advanced Planning agent that decomposes auxiliary tasks into step-by-step "
                "plans for another advanced agent, Executor."
            )

        return ChatPromptTemplate.from_messages(
            [
                ("system", system_message),
                (
                    "human",
                    user_message,
                ),
                (
                    "human",
                    "Here is the final outcome of the original task:",
                ),
                MessagesPlaceholder("executor_agent_outcome"),
                (
                    "human",
                    "Below, you will find the intermediate steps that were taken in order to achieve this outcome.",
                ),
                MessagesPlaceholder("executor_intermediate_steps"),
                (
                    "human",
                    dedent("""
                            The trial above was unsuccessful. Your goal is to construct a step-by-step plan to successfully solve the original task.
                            Plan is a list of subtasks with the logic of how the subtasks' results should be aggregated.

                            You are only allowed to devise plans either with "and" aggregation logic or "or" aggregation logic.
                            For "and" logic, the original task will only be considered solved if ALL subtasks are successfully solved.
                            For "or" logic, the original task will only be considered solved if ANY of the subtasks are successfully solved.
                            In both cases, subtasks will be executed sequentially. 
                            For "and" logic, the execution will stop at the first failure.
                            For "or" logic, the execution will stop at the first success.

                            Each subtask will be passed to an executor agent separately. Make sure to make all the instructions self-sufficient (containing all the necessary information, observations, etc.), yet concise.

                            Use JSON format to output a plan. It should contain keys 'subtasks' (list of subtasks, where each subtask is defined by a simple natural-language instruction, a single string) and 'aggregation_mode' (logic for uniting the subtasks results), like that:
                            ```
                            {{"subtasks": [
                                          "<subtask1 instruction>", 
                                          "<subtask2 instruction>", 
                                          ..., 
                                          "<subtaskn instruction>"],
                             "aggregation_mode": "<aggregation mode>"}}
                            ```    
                            """),
                ),
            ]
        )

    def invoke(
        self,
        inputs: ADaPTPlannerInput,
        run_manager: Optional[CallbackManager] = None,
        **kwargs,
    ) -> ADaPTPlannerOutput:
        return self.runnable.invoke(inputs, run_manager=run_manager, run_name=self.name)

    async def ainvoke(
        self,
        inputs: ADaPTPlannerInput,
        run_manager: Optional[AsyncCallbackManager] = None,
        **kwargs,
    ) -> ADaPTPlannerOutput:
        return await self.runnable.ainvoke(inputs, run_manager=run_manager, run_name=self.name)

    @classmethod
    def create(
        cls,
        llm: BaseChatModel,
        prompt: Optional[ChatPromptTemplate] = None,
        user_message: Optional[str] = None,
        system_message: Optional[str] = None,
        executor_parser: Optional[
            Union[
                BaseFunctionCallingSingleActionParser,
                BaseFunctionCallingMultiActionParser,
            ]
        ] = None,
        executor_parser_name: Optional[str] = None,
        output_parser=None,
        **kwargs,
    ) -> "ADaPTPlanner":
        def _preprocess_input(
            inputs: Dict[str, Any],
        ) -> Dict[str, Any]:
            # TODO: figure out typing here
            nonlocal executor_parser, executor_parser_name
            if executor_parser is None:
                assert executor_parser_name is not None
                executor_parser = ParserRegistry.get_parser(executor_parser_name)

            executor_intermediate_steps = executor_parser.format_inputs(
                {
                    "inputs": inputs["inputs"],
                    "intermediate_steps": inputs["executor_intermediate_steps"],
                }
            )["agent_scratchpad"]

            return {
                **inputs["inputs"],
                "executor_agent_outcome": format_thought(inputs["executor_agent_outcome"]),
                "executor_intermediate_steps": executor_intermediate_steps,
            }

        prompt = cls._process_prompt(
            prompt=prompt,
            user_message=user_message,
            system_message=system_message,
            mode="simple",
        )

        if output_parser is None:
            output_parser = SimplePlannerOutputParser()

        runnable = RunnableComponent.create_from_steps(prompt=prompt, llm=llm, output_parser=output_parser)
        runnable.add_input_preprocessing(_preprocess_input)
        return cls(runnable=runnable)



from __future__ import annotations

from typing import Any, Dict, Generic, Optional, Tuple, TypeVar

import gymnasium as gym
from langchain.pydantic_v1 import BaseModel, Field
from langchain.tools import BaseTool
from langchain_core.agents import AgentAction
from langchain_core.callbacks import CallbackManager

ObsType = TypeVar("ObsType")


class GymEnvResetTool(BaseTool, BaseModel, Generic[ObsType]):
    env: gym.Env[ObsType, Tuple[AgentAction, Optional[CallbackManager]]] = Field(  # type: ignore[valid-type]
        exclude=True
    )

    name: str = "reset"
    description: str = "Resets the environment state."

    class Config(BaseTool.Config):
        pass

    def _run(
        self,
        *,
        seed: int | None = None,
        options: Dict[str, Any] | None = None,
    ) -> Tuple[ObsType, Dict[str, Any]]:
        return self.env.reset(seed=seed, options=options)  # type: ignore[reportReturnType]



from .base_action_executor import BaseActionExecutor
from .default_action_executor import LangchainActionExecutor
from .meta_tools import MetaTools

__all__ = ["BaseActionExecutor", "LangchainActionExecutor", "MetaTools"]



from __future__ import annotations

from typing import AsyncIterator, Dict, Iterator, List, Optional, Sequence, Tuple, Union

from langchain.agents import BaseMultiActionAgent, BaseSingleActionAgent
from langchain_core.agents import AgentAction, AgentFinish, AgentStep
from langchain_core.callbacks import (
    AsyncCallbackManagerForChainRun,
    CallbackManagerForChainRun,
)
from langchain_core.tools import BaseTool

from planning_library.action_executors import (
    BaseActionExecutor,
    LangchainActionExecutor,
    MetaTools,
)

from ..base_strategy import BaseCustomStrategy


class SimpleStrategy(BaseCustomStrategy):
    """Simple strategy akin to langchain.agents.AgentExecutor:
    calls agent in a loop until either AgentFinish is produced or early stopping condition in reached."""

    action_executor: BaseActionExecutor
    agent: BaseSingleActionAgent | BaseMultiActionAgent

    @property
    def input_keys(self) -> List[str]:
        return self.agent.input_keys

    @property
    def output_keys(self) -> List[str]:
        if self.return_intermediate_steps:
            return self.agent.return_values + ["intermediate_steps"]
        else:
            return self.agent.return_values

    @classmethod
    def create(
        cls,
        meta_tools: Optional[MetaTools] = None,
        return_intermediate_steps: bool = False,
        return_finish_log: bool = False,
        max_iterations: int = 15,
        verbose: bool = True,
        action_executor: Optional[BaseActionExecutor] = None,
        tools: Optional[Sequence[BaseTool]] = None,
        agent: Optional[Union[BaseSingleActionAgent, BaseMultiActionAgent]] = None,
        **kwargs,
    ) -> "SimpleStrategy":
        tools = tools if tools is not None else []
        action_executor = (
            action_executor
            if action_executor is not None
            else LangchainActionExecutor(tools=tools, meta_tools=meta_tools)
        )

        if agent is None:
            raise ValueError("Default agent is currently not supported.")

        return SimpleStrategy(
            agent=agent,
            action_executor=action_executor,
            return_intermediate_steps=return_intermediate_steps,
            return_finish_log=return_finish_log,
            max_iterations=max_iterations,
            verbose=verbose,
        )

    def _run_strategy(
        self,
        inputs: Dict[str, str],
        run_manager: Optional[CallbackManagerForChainRun] = None,
    ) -> Iterator[Tuple[AgentFinish, List[Tuple[AgentAction, str]]]]:
        intermediate_steps: List[Tuple[AgentAction, str]] = []

        cur_iteration = 0
        while self.max_iterations is None or cur_iteration < self.max_iterations:
            agent_outcome = self.agent.plan(
                intermediate_steps,
                callbacks=run_manager.get_child() if run_manager else None,
                **inputs,
            )

            if isinstance(agent_outcome, AgentFinish):
                yield agent_outcome, intermediate_steps
                return

            action_results = self.action_executor.execute(
                agent_outcome,
                run_manager=run_manager.get_child() if run_manager else None,
            )

            if isinstance(action_results, AgentStep):
                intermediate_steps.append((action_results.action, action_results.observation))
            else:
                intermediate_steps.extend(
                    (_action_results.action, _action_results.observation) for _action_results in action_results
                )

            cur_iteration += 1

        stopped_outcome = AgentFinish({"output": "Agent stopped due to iteration limit."}, "")

        yield stopped_outcome, intermediate_steps
        return

    async def _arun_strategy(
        self,
        inputs: Dict[str, str],
        run_manager: Optional[AsyncCallbackManagerForChainRun] = None,
    ) -> AsyncIterator[Tuple[AgentFinish, List[Tuple[AgentAction, str]]]]:
        intermediate_steps: List[Tuple[AgentAction, str]] = []

        cur_iteration = 0
        while self.max_iterations is None or cur_iteration < self.max_iterations:
            agent_outcome = await self.agent.aplan(
                intermediate_steps,
                callbacks=run_manager.get_child() if run_manager else None,
                **inputs,
            )

            if isinstance(agent_outcome, AgentFinish):
                yield agent_outcome, intermediate_steps
                return

            action_results = await self.action_executor.aexecute(agent_outcome)

            if isinstance(action_results, AgentStep):
                intermediate_steps.append((action_results.action, action_results.observation))
            else:
                intermediate_steps.extend(
                    (_action_results.action, _action_results.observation) for _action_results in action_results
                )

            cur_iteration += 1

        stopped_outcome = AgentFinish({"output": "Agent stopped due to iteration limit."}, "")

        yield stopped_outcome, intermediate_steps
        return



from functools import partial
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    TypedDict,
    Union,
)

from langchain.memory import ChatMessageHistory
from langchain_core.agents import AgentAction, AgentFinish, AgentStep
from langchain_core.chat_history import BaseChatMessageHistory
from langchain_core.messages import AIMessage, BaseMessage
from langchain_core.runnables import RunnableLambda
from langgraph.graph import END, StateGraph  # type: ignore[import]
from langgraph.pregel import Pregel  # type: ignore[import-untyped]

from ...action_executors import BaseActionExecutor
from .components import (
    ReflexionActor,
    ReflexionActorInput,
    ReflexionEvaluator,
    ReflexionEvaluatorInput,
    ReflexionSelfReflection,
    ReflexionSelfReflectionInput,
)


class ReflexionState(TypedDict):
    """A state passed between nodes of Reflexion graph."""

    inputs: Dict[str, Any]
    agent_outcome: Optional[Union[List[AgentAction], AgentAction, AgentFinish]]
    evaluator_should_continue: Optional[bool]
    self_reflection_memory: BaseChatMessageHistory
    self_reflections: Sequence[BaseMessage]
    intermediate_steps: List[Tuple[AgentAction, str]]
    iteration: int


class ReflexionNodes:
    @staticmethod
    def _format_self_reflections(
        self_reflections: List[Tuple[str, str]],
    ) -> Sequence[BaseMessage]:
        result = []
        for t in self_reflections:
            if t[0] == "content":
                content = t[1]
                message = AIMessage(content=content)
                result.append(message)

        return result

    @staticmethod
    def init(state: ReflexionState, memory: Optional[BaseChatMessageHistory] = None) -> ReflexionState:
        """The entry node in the graph. Initializes the state correctly."""
        state["agent_outcome"] = None
        state["evaluator_should_continue"] = None
        state["self_reflection_memory"] = ChatMessageHistory() if memory is None else memory
        state["self_reflections"] = []
        state["intermediate_steps"] = []
        state["iteration"] = 1
        return state

    @staticmethod
    def re_init(
        state: ReflexionState,
        reset_environment: Optional[Callable[[Dict[str, Any]], None]],
    ) -> ReflexionState:
        """The first node that gets called after at least one iteration. Handles the advance of the loop interation correctly."""
        state["agent_outcome"] = None
        state["evaluator_should_continue"] = None
        state["intermediate_steps"] = []
        state["iteration"] += 1
        # TODO: why does memory return list of tuples instead of messages as expected? some serialization stuff?
        state["self_reflections"] = ReflexionNodes._format_self_reflections(
            state["self_reflection_memory"].messages  # type: ignore[arg-type]
        )

        if reset_environment:
            reset_environment(state["inputs"])

        return state

    @staticmethod
    def act(state: ReflexionState, actor: ReflexionActor) -> ReflexionState:
        """Synchronous version of calling an agent and returning its result."""
        agent_outcome = actor.invoke(
            ReflexionActorInput(
                inputs=state["inputs"],
                intermediate_steps=state["intermediate_steps"],
                self_reflections=state["self_reflections"],
            )
        )
        state["agent_outcome"] = agent_outcome
        return state

    @staticmethod
    async def aact(state: ReflexionState, actor: ReflexionActor) -> ReflexionState:
        """Asynchronous version of calling an agent and returning its result."""
        agent_outcome = await actor.ainvoke(
            ReflexionActorInput(
                inputs=state["inputs"],
                intermediate_steps=state["intermediate_steps"],
                self_reflections=state["self_reflections"],
            )
        )

        state["agent_outcome"] = agent_outcome
        return state

    @staticmethod
    def execute_actions(
        state: ReflexionState,
        action_executor: BaseActionExecutor,
    ) -> ReflexionState:
        """Synchronous version of executing actions as previously requested by an agent."""
        assert state["agent_outcome"] is not None, "Agent outcome should be defined on the tool execution step."
        assert not isinstance(
            state["agent_outcome"], AgentFinish
        ), "Agent outcome should not be AgentFinish on the tool execution step."

        observation = action_executor.execute(
            actions=state["agent_outcome"],
        )

        if isinstance(observation, AgentStep):
            state["intermediate_steps"].append((observation.action, observation.observation))
        else:
            state["intermediate_steps"].extend([(obs.action, obs.observation) for obs in observation])
        return state

    @staticmethod
    async def aexecute_actions(
        state: ReflexionState,
        action_executor: BaseActionExecutor,
    ) -> ReflexionState:
        """Asynchronous version of executing tools as previously requested by an agent."""
        assert state["agent_outcome"] is not None, "Agent outcome should be defined on the tool execution step."
        assert not isinstance(
            state["agent_outcome"], AgentFinish
        ), "Agent outcome should not be AgentFinish on the tool execution step."

        observation = await action_executor.aexecute(
            actions=state["agent_outcome"],
        )

        if isinstance(observation, AgentStep):
            state["intermediate_steps"].append((observation.action, observation.observation))
        else:
            state["intermediate_steps"].extend([(obs.action, obs.observation) for obs in observation])
        return state

    @staticmethod
    def evaluate(state: ReflexionState, evaluator: ReflexionEvaluator) -> ReflexionState:
        """Synchronous version of evaluating the outcome of the current trial."""
        assert isinstance(
            state["agent_outcome"], AgentFinish
        ), "Agent outcome should be AgentFinish on the evaluation step."
        should_continue = evaluator.invoke(
            ReflexionEvaluatorInput(
                inputs=state["inputs"],
                intermediate_steps=state["intermediate_steps"],
                agent_outcome=state["agent_outcome"],
            )
        )
        state["evaluator_should_continue"] = should_continue
        return state

    @staticmethod
    async def aevaluate(state: ReflexionState, evaluator: ReflexionEvaluator) -> ReflexionState:
        """Asynchronous version of evaluating the outcome of the current trial."""
        assert isinstance(
            state["agent_outcome"], AgentFinish
        ), "Agent outcome should be AgentFinish on the evaluation step."

        should_continue = await evaluator.ainvoke(
            ReflexionEvaluatorInput(
                inputs=state["inputs"],
                intermediate_steps=state["intermediate_steps"],
                agent_outcome=state["agent_outcome"],
            )
        )
        state["evaluator_should_continue"] = should_continue
        return state

    @staticmethod
    def self_reflect(state: ReflexionState, self_reflection: ReflexionSelfReflection) -> ReflexionState:
        """Synchronous version of self-reflecting on the current trial."""
        assert isinstance(
            state["agent_outcome"], AgentFinish
        ), "Agent outcome should be AgentFinish on the self-reflection step."

        reflection = self_reflection.invoke(
            ReflexionSelfReflectionInput(
                inputs=state["inputs"],
                intermediate_steps=state["intermediate_steps"],
                agent_outcome=state["agent_outcome"],
            )
        )
        state["self_reflection_memory"].add_messages(reflection)
        return state

    @staticmethod
    async def aself_reflect(state: ReflexionState, self_reflection: ReflexionSelfReflection) -> ReflexionState:
        """Asynchronous version of self-reflecting on the current trial."""
        assert isinstance(
            state["agent_outcome"], AgentFinish
        ), "Agent outcome should be AgentFinish on the self-reflection step."

        reflection = await self_reflection.ainvoke(
            ReflexionSelfReflectionInput(
                inputs=state["inputs"],
                intermediate_steps=state["intermediate_steps"],
                agent_outcome=state["agent_outcome"],
            )
        )
        await state["self_reflection_memory"].aadd_messages(reflection)
        return state


class ReflexionEdges:
    @staticmethod
    def should_continue_evaluator(state: ReflexionState) -> Literal["yes", "no"]:
        """Conditional edge that determines whether the main loop should be continued or stopped based on evaluator's output.

        If yes: execution finishes.
        If no: next iteration.
        """
        return "yes" if state["evaluator_should_continue"] else "no"

    @staticmethod
    def should_continue_actor(state: ReflexionState) -> Literal["yes", "no"]:
        """Conditional edge that determines whether the agent loop should be continued or stopped based on agent's output.

        If yes: proceed to evaluation.
        If no: proceed to act.
        """
        return "no" if isinstance(state["agent_outcome"], AgentFinish) else "yes"

    @staticmethod
    def should_continue_num_iterations(
        state: ReflexionState,
        max_iterations: Optional[int],
    ) -> Literal["yes", "no"]:
        """Conditional edge that determines whether the main loop should be continued or stopped based on the number of iterations threshold.

        If yes: execution finishes.
        If no: next iteration.
        """
        if max_iterations is not None and state["iteration"] >= max_iterations:
            return "no"
        return "yes"


def create_reflexion_graph(
    actor: ReflexionActor,
    evaluator: ReflexionEvaluator,
    self_reflection: ReflexionSelfReflection,
    action_executor: BaseActionExecutor,
    max_iterations: Optional[int],
    reset_environment: Optional[Callable[[Dict[str, Any]], None]],
) -> Pregel:
    """Builds a graph for Reflexion strategy."""

    builder = StateGraph(ReflexionState)
    builder.add_node("init", ReflexionNodes.init)
    builder.add_node(
        "re_init",
        partial(ReflexionNodes.re_init, reset_environment=reset_environment),
    )
    builder.add_node(
        "act",
        RunnableLambda(
            partial(ReflexionNodes.act, actor=actor),
            afunc=partial(ReflexionNodes.aact, actor=actor),
        ),
    )

    builder.add_node(
        "execute_actions",
        RunnableLambda(
            partial(
                ReflexionNodes.execute_actions,
                action_executor=action_executor,
            ),
            afunc=partial(
                ReflexionNodes.aexecute_actions,
                action_executor=action_executor,
            ),
        ),
    )
    builder.add_node(
        "evaluate",
        RunnableLambda(
            partial(ReflexionNodes.evaluate, evaluator=evaluator),
            afunc=partial(ReflexionNodes.aevaluate, evaluator=evaluator),
        ),
    )
    builder.add_node(
        "self_reflect",
        RunnableLambda(
            partial(ReflexionNodes.self_reflect, self_reflection=self_reflection),
            afunc=partial(ReflexionNodes.aself_reflect, self_reflection=self_reflection),
        ),
    )

    builder.set_entry_point("init")
    builder.add_edge("init", "act")

    builder.add_conditional_edges(
        "act",
        ReflexionEdges.should_continue_actor,
        {"yes": "execute_actions", "no": "evaluate"},
    )
    builder.add_edge("execute_actions", "act")
    builder.add_conditional_edges(
        "evaluate",
        ReflexionEdges.should_continue_evaluator,
        {
            "yes": "self_reflect",
            "no": END,
        },
    )
    builder.add_conditional_edges(
        "self_reflect",
        partial(ReflexionEdges.should_continue_num_iterations, max_iterations=max_iterations),
        {"yes": "re_init", "no": END},
    )
    builder.add_edge("re_init", "act")
    return builder.compile()



from typing import Optional

from langchain_core.callbacks import AsyncCallbackManager, CallbackManager

from planning_library.components.base_component import InputType

from ..base_component import BaseComponent


class LeqThresholdJudge(BaseComponent[InputType, bool]):
    def __init__(self, threshold: float):
        self.threshold = threshold

    def invoke(self, inputs: InputType, run_manager: Optional[CallbackManager] = None, **kwargs) -> bool:
        return inputs["backbone_output"] <= self.threshold

    async def ainvoke(
        self,
        inputs: InputType,
        run_manager: Optional[AsyncCallbackManager] = None,
        **kwargs,
    ) -> bool:
        return inputs["backbone_output"] <= self.threshold


class GeqThresholdJudge(BaseComponent[InputType, bool]):
    def __init__(self, threshold: float):
        self.threshold = threshold

    def invoke(self, inputs: InputType, run_manager: Optional[CallbackManager] = None, **kwargs) -> bool:
        return inputs["backbone_output"] >= self.threshold

    async def ainvoke(
        self,
        inputs: InputType,
        run_manager: Optional[AsyncCallbackManager] = None,
        **kwargs,
    ) -> bool:
        return inputs["backbone_output"] >= self.threshold



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from itertools import combinations
from textwrap import dedent
from typing import Any, Dict, List, Optional, Tuple, Union

from langchain_core.agents import AgentAction, AgentFinish
from langchain_core.callbacks import AsyncCallbackManager, CallbackManager
from langchain_core.language_models import BaseChatModel
from langchain_core.messages import BaseMessage
from langchain_core.output_parsers import BaseOutputParser
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain_core.runnables import Runnable
from typing_extensions import TypedDict

from planning_library.components import BaseComponent, RunnableComponent
from planning_library.function_calling_parsers import (
    BaseFunctionCallingMultiActionParser,
    BaseFunctionCallingSingleActionParser,
    ParserRegistry,
)
from planning_library.utils import (
    format_thought,
)


@dataclass
class ThoughtSorterConfig:
    runnable: Optional[Runnable] = None

    prompt: Optional[ChatPromptTemplate] = None
    user_message: Optional[str] = None
    system_message: Optional[str] = None

    llm: Optional[BaseChatModel] = None

    parser: Optional[
        Union[
            BaseFunctionCallingSingleActionParser,
            BaseFunctionCallingMultiActionParser,
        ]
    ] = None
    parser_name: Optional[str] = None

    output_parser: Optional[BaseOutputParser[str]] = None


class ThoughtSorterInput(TypedDict):
    inputs: Dict[str, Any]
    thoughts: List[List[AgentAction] | AgentAction | AgentFinish]
    intermediate_steps: List[Tuple[AgentAction, str]]


class ThoughtSorterRunnableInput(TypedDict):
    intermediate_steps: List[Tuple[AgentAction, str]]
    thought1: List[BaseMessage]
    thought2: List[BaseMessage]


class ThoughtSorter(BaseComponent[ThoughtSorterInput, List[Union[List[AgentAction], AgentAction, AgentFinish]]]):
    """
    ToT+DFS component responsible for sorting the candidate thought on each DFS step.

    Follows the algorithm from ToolLLM repository (https://github.com/OpenBMB/ToolBench), specifically:
      1. Compares all pairs of candidate thoughts
      2. Computes final scores for each thought based on pairwise comparison
      3. Returns a list sorted by the scores (in descending order, the bigger the better).

    https://github.com/OpenBMB/ToolBench/blob/2937497244096960a532b21f66f663ed78e08588/toolbench/inference/LLM_rank/rank_candidate.py#L53
    """

    name = "Sort Thoughts"

    def __init__(
        self,
        runnable: Runnable[ThoughtSorterRunnableInput, str] | RunnableComponent[ThoughtSorterRunnableInput, str],
    ):
        if not isinstance(runnable, RunnableComponent):
            runnable = RunnableComponent(runnable)
        self.runnable = runnable

    @classmethod
    def _create_default_prompt(cls, system_message: Optional[str], user_message: str, **kwargs) -> ChatPromptTemplate:
        if system_message is None:
            system_message = (
                "You are an advanced reasoning assistant that compares the "
                "steps suggested for solving complex reasoning tasks."
            )

        return ChatPromptTemplate.from_messages(
            [
                ("system", system_message),
                (
                    "human",
                    user_message,
                ),
                MessagesPlaceholder("intermediate_steps"),
                ("human", "Here is the first proposed next step"),
                MessagesPlaceholder("thought1"),
                ("human", "Here is the second proposed next step:"),
                MessagesPlaceholder("thought2"),
                (
                    "human",
                    dedent("""Your goal is to judge which of the proposed actions is more likely to lead to the success.

                         Take your time and comment your decision, 
                         but make sure to always output either 1 or 2, 
                         where 1 means 'the first proposed action is more likely to help' 
                         and 2 means 'the second proposed action is more likely to help'. 

                         ALWAYS use the following format and add the number in the end of your answer: [[number]].

                         Your verdict:
                         """),
                ),
            ]
        )

    def _compare_pairwise(
        self,
        inputs: Dict[str, Any],
        intermediate_steps: List[Tuple[AgentAction, str]],
        thought1: List[AgentAction] | AgentAction | AgentFinish,
        thought2: List[AgentAction] | AgentAction | AgentFinish,
        run_manager: Optional[CallbackManager] = None,
        **kwargs,
    ) -> str:
        return self.runnable.invoke(
            {
                **inputs,  # type: ignore[typeddict-item]
                "intermediate_steps": intermediate_steps,
                "thought1": format_thought(thought1),
                "thought2": format_thought(thought2),
            },
            run_manager=run_manager,
            **kwargs,
        )

    async def _acompare_pairwise(
        self,
        inputs: Dict[str, Any],
        intermediate_steps: List[Tuple[AgentAction, str]],
        thought1: List[AgentAction] | AgentAction | AgentFinish,
        thought2: List[AgentAction] | AgentAction | AgentFinish,
        run_manager: Optional[AsyncCallbackManager] = None,
        **kwargs,
    ) -> str:
        return await self.runnable.ainvoke(
            {
                **inputs,  # type: ignore[typeddict-item]
                "intermediate_steps": intermediate_steps,
                "thought1": format_thought(thought1),
                "thought2": format_thought(thought2),
            },
            run_manager=run_manager,
        )

    def invoke(
        self,
        inputs: ThoughtSorterInput,
        run_manager: Optional[CallbackManager] = None,
        **kwargs,
    ) -> List[Union[List[AgentAction], AgentAction, AgentFinish]]:
        scores: Dict[Union[List[AgentAction], AgentAction, AgentFinish], float] = defaultdict(float)
        for thought1, thought2 in [pair for pair in combinations(inputs["thoughts"], 2)]:
            result = self._compare_pairwise(
                inputs=inputs["inputs"],
                intermediate_steps=inputs["intermediate_steps"],
                thought1=thought1,
                thought2=thought2,
                run_manager=run_manager,
                **kwargs,
            )

            if result == "1":
                scores[thought1] += 1
            elif result == "2":
                scores[thought2] += 1
            else:
                scores[thought1] += 0.5
                scores[thought2] += 0.5

        sorted_scores = sorted(scores, key=lambda x: scores[x], reverse=True)
        return [key for key in sorted_scores]

    async def ainvoke(
        self,
        inputs: ThoughtSorterInput,
        run_manager: Optional[AsyncCallbackManager] = None,
        **kwargs,
    ) -> List[Union[List[AgentAction], AgentAction, AgentFinish]]:
        scores: Dict[Union[List[AgentAction], AgentAction, AgentFinish], float] = defaultdict(float)
        for thought1, thought2 in [pair for pair in combinations(inputs["thoughts"], 2)]:
            result = await self._acompare_pairwise(
                inputs=inputs["inputs"],
                intermediate_steps=inputs["intermediate_steps"],
                thought1=thought1,
                thought2=thought2,
                run_manager=run_manager,
            )

            if result == "1":
                scores[thought1] += 1
            elif result == "2":
                scores[thought2] += 1
            else:
                scores[thought1] += 0.5
                scores[thought2] += 0.5

        sorted_scores = sorted(scores, key=lambda x: scores[x], reverse=True)
        return [key for key in sorted_scores]

    @classmethod
    def create_from_config(cls, config: ThoughtSorterConfig) -> ThoughtSorter:
        def _preprocess_input(
            inputs: ThoughtSorterInput,
        ) -> Dict:
            # TODO: figure out typing here
            nonlocal config
            if config.parser is None:
                assert config.parser_name is not None
                parser = ParserRegistry.get_parser(config.parser_name)
            else:
                parser = config.parser

            intermediate_steps = parser.format_inputs(inputs)["agent_scratchpad"]
            return {
                **inputs["inputs"],
                "thoughts": inputs["thoughts"],
                "intermediate_steps": intermediate_steps,
            }

        if config.runnable is not None:
            return cls(config.runnable)

        if config.llm is None:
            raise ValueError("`llm` must be provided when `runnable` is None.")

        if config.output_parser is None:
            raise ValueError(
                "Default output parser for thought sorter is not implemented yet.`output_parser` must be provided when `runnable` is None."
            )

        prompt = cls._process_prompt(
            prompt=config.prompt,
            user_message=config.user_message,
            system_message=config.system_message,
        )

        sorter_runnable = RunnableComponent.create_from_steps(
            prompt=prompt, llm=config.llm, output_parser=config.output_parser
        )

        sorter_runnable.add_input_preprocessing(_preprocess_input)

        return cls(sorter_runnable)



from typing import Any, Callable, Dict, Optional

from ...action_executors import BaseActionExecutor
from ..base_strategy import BaseLangGraphStrategy
from .components import ReflexionActor, ReflexionEvaluator, ReflexionSelfReflection
from .reflexion_graph import create_reflexion_graph


class ReflexionStrategy(BaseLangGraphStrategy):
    """Reflexion strategy.

    Based on "Reflexion: Language Agents with Verbal Reinforcement Learning", Shinn et al.
    and an example in LangGraph repo: https://github.com/langchain-ai/langgraph/tree/main/examples/reflexion
    """

    @staticmethod
    def create_from_components(
        actor: ReflexionActor,
        evaluator: ReflexionEvaluator,
        self_reflection: ReflexionSelfReflection,
        action_executor: BaseActionExecutor,
        max_iterations: int,
        reset_environment: Optional[Callable[[Dict[str, Any]], None]] = None,
    ):
        return create_reflexion_graph(
            actor=actor,
            evaluator=evaluator,
            self_reflection=self_reflection,
            action_executor=action_executor,
            max_iterations=max_iterations,
            reset_environment=reset_environment,
        )



from typing import List, Union

from langchain.agents.agent import RunnableAgent, RunnableMultiActionAgent
from langchain_core.agents import AgentAction, AgentFinish
from langchain_core.runnables import Runnable


def convert_runnable_to_agent(
    agent: Runnable,
) -> Union[RunnableAgent, RunnableMultiActionAgent]:
    """Convert runnable to agent if passed in.

    Copied from langchain.agents.agent.AgentExecutor.validate_runnable_agent.
    """
    try:
        output_type = agent.OutputType
    except Exception as _:
        multi_action = False
    else:
        multi_action = output_type == Union[List[AgentAction], AgentFinish]

    if multi_action:
        return RunnableMultiActionAgent(runnable=agent)
    return RunnableAgent(runnable=agent)



from langchain_core.output_parsers import BaseOutputParser
from langchain_core.utils.json import parse_and_check_json_markdown

from .typing_utils import ADaPTPlannerOutput


class SimplePlannerOutputParser(BaseOutputParser[ADaPTPlannerOutput]):
    def parse(self, text: str) -> ADaPTPlannerOutput:
        output = parse_and_check_json_markdown(text, expected_keys=["subtasks", "aggregation_mode"])
        return {
            "subtasks": output["subtasks"],
            "aggregation_mode": output["aggregation_mode"],
        }



from .actor import ReflexionActor, ReflexionActorInput
from .evaluator import ReflexionEvaluator, ReflexionEvaluatorInput
from .self_reflection import ReflexionSelfReflection, ReflexionSelfReflectionInput

__all__ = [
    "ReflexionActor",
    "ReflexionActorInput",
    "ReflexionEvaluator",
    "ReflexionEvaluatorInput",
    "ReflexionSelfReflection",
    "ReflexionSelfReflectionInput",
]



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

from langchain.agents import BaseMultiActionAgent, BaseSingleActionAgent
from langchain_core.agents import AgentAction, AgentFinish
from langchain_core.callbacks import AsyncCallbackManager, CallbackManager
from langchain_core.language_models import BaseChatModel
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain_core.tools import BaseTool
from typing_extensions import TypedDict

from planning_library.components import AgentComponent, BaseComponent
from planning_library.function_calling_parsers import (
    BaseFunctionCallingMultiActionParser,
    BaseFunctionCallingSingleActionParser,
)
from planning_library.utils import (
    format_thoughts,
)


@dataclass
class ThoughtGeneratorConfig:
    tools: Sequence[BaseTool]
    max_num_thoughts: int

    prompt: Optional[ChatPromptTemplate] = None
    user_message: Optional[str] = None
    system_message: Optional[str] = None

    agent: Optional[BaseMultiActionAgent | BaseSingleActionAgent] = None

    llm: Optional[BaseChatModel] = None

    parser: Optional[
        Union[
            BaseFunctionCallingSingleActionParser,
            BaseFunctionCallingMultiActionParser,
        ]
    ] = None
    parser_name: Optional[str] = None


class ThoughtGeneratorInput(TypedDict):
    inputs: Dict[str, Any]
    intermediate_steps: List[Tuple[AgentAction, str]]


class ThoughtGeneratorAgentInput(ThoughtGeneratorInput):
    previous_thoughts: List[List[AgentAction] | AgentAction | AgentFinish]


class ThoughtGenerator(BaseComponent[ThoughtGeneratorInput, List[Union[List[AgentAction], AgentAction, AgentFinish]]]):
    name = "Generate Thoughts"

    required_prompt_input_vars = set(ThoughtGeneratorInput.__annotations__) - {
        "inputs",
        "intermediate_steps",
    } | {"agent_scratchpad"}

    def __init__(
        self,
        agent: AgentComponent | BaseMultiActionAgent | BaseSingleActionAgent,
        max_num_thoughts: int,
    ):
        self.agent: AgentComponent[ThoughtGeneratorAgentInput] = (
            AgentComponent(agent) if not isinstance(agent, AgentComponent) else agent
        )
        self.max_num_thoughts = max_num_thoughts

    @classmethod
    def _create_default_prompt(cls, system_message: Optional[str], user_message: str, **kwargs) -> ChatPromptTemplate:
        if system_message is None:
            system_message = "You are an advanced reasoning agent that can improve based on self-reflection."

        return ChatPromptTemplate.from_messages(
            [
                ("system", system_message),
                (
                    "human",
                    user_message,
                ),
                MessagesPlaceholder("agent_scratchpad"),
                (
                    "human",
                    "You might have already made some suggestions for the current state - if you did, you will find them below.",
                ),
                MessagesPlaceholder("previous_thoughts"),
                (
                    "human",
                    "Please, remember to suggest exactly ONE (1) tool call, no more and no less, different from your previous suggestions.",
                ),
            ]
        )

    def invoke(
        self,
        inputs: ThoughtGeneratorInput,
        run_manager: Optional[CallbackManager] = None,
        **kwargs,
    ) -> List[List[AgentAction] | AgentAction | AgentFinish]:
        results: List[List[AgentAction] | AgentAction | AgentFinish] = []
        for _ in range(self.max_num_thoughts):
            cur_result = self.agent.invoke(
                {**inputs, "previous_thoughts": results},
                run_manager=run_manager,
                **kwargs,
            )
            # TODO: how to fix mypy warning properly here?
            results.append(cur_result)  # type: ignore[arg-type]

        return results

    async def ainvoke(
        self,
        inputs: ThoughtGeneratorInput,
        run_manager: Optional[AsyncCallbackManager] = None,
        **kwargs,
    ) -> List[List[AgentAction] | AgentAction | AgentFinish]:
        results: List[List[AgentAction] | AgentAction | AgentFinish] = []
        for _ in range(self.max_num_thoughts):
            cur_result = await self.agent.ainvoke(
                {**inputs, "previous_thoughts": results},
                run_manager=run_manager,
                **kwargs,
            )
            # TODO: how to fix mypy warning properly here?
            results.append(cur_result)  # type: ignore[arg-type]

        return results

    @classmethod
    def create_from_config(cls, config: ThoughtGeneratorConfig) -> ThoughtGenerator:
        if config.agent is not None:
            return ThoughtGenerator(agent=config.agent, max_num_thoughts=config.max_num_thoughts)

        if config.llm is None:
            raise ValueError("`llm` must be provided when `agent` is None.")

        if config.tools is None:
            raise ValueError("`tools` must be provided when `agent` is None.")

        return cls.create(
            llm=config.llm,
            tools=config.tools,
            prompt=config.prompt,
            user_message=config.user_message,
            system_message=config.system_message,
            parser=config.parser,
            parser_name=config.parser_name,
            max_num_thoughts=config.max_num_thoughts,
        )

    @classmethod
    def create(
        cls,
        llm: BaseChatModel,
        tools: Sequence[BaseTool],
        max_num_thoughts: int,
        prompt: Optional[ChatPromptTemplate] = None,
        user_message: Optional[str] = None,
        system_message: Optional[str] = None,
        parser: Optional[
            Union[
                BaseFunctionCallingSingleActionParser,
                BaseFunctionCallingMultiActionParser,
            ]
        ] = None,
        parser_name: Optional[str] = None,
    ) -> ThoughtGenerator:
        prompt = cls._process_prompt(prompt=prompt, user_message=user_message, system_message=system_message)

        agent: AgentComponent = AgentComponent.create(
            llm=llm,
            tools=tools,
            prompt=prompt,
            parser=parser,
            parser_name=parser_name,
        )

        agent.add_input_preprocessing(
            preprocess=lambda inputs: {
                **{key: value for key, value in inputs.items() if key != "previous_thoughts"},
                "previous_thoughts": format_thoughts(inputs["previous_thoughts"]),
            }
        )

        return ThoughtGenerator(agent=agent, max_num_thoughts=max_num_thoughts)



from __future__ import annotations

from textwrap import dedent
from typing import Any, Dict, Generic, List, Optional, Tuple, Type

from langchain_core.agents import AgentAction, AgentFinish
from langchain_core.language_models import BaseChatModel
from langchain_core.messages import BaseMessage
from langchain_core.output_parsers import BaseOutputParser
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from typing_extensions import TypedDict

from planning_library.components.base_component import OutputType
from planning_library.components.evaluation import EvaluatorComponent
from planning_library.function_calling_parsers import (
    BaseFunctionCallingMultiActionParser,
    BaseFunctionCallingSingleActionParser,
    ParserRegistry,
)


class ReflexionEvaluatorInput(TypedDict):
    inputs: Dict[str, Any]
    intermediate_steps: List[Tuple[AgentAction, str]]
    agent_outcome: AgentFinish


class PreprocessedReflexionEvaluatorInput(TypedDict):
    intermediate_steps: List[BaseMessage]
    agent_outcome: str


class ReflexionEvaluator(Generic[OutputType], EvaluatorComponent[ReflexionEvaluatorInput, OutputType]):
    name = "Evaluator"

    required_prompt_input_vars = set(ReflexionEvaluatorInput.__annotations__) - {"inputs"}

    @classmethod
    def _create_default_prompt(cls, system_message: Optional[str], user_message: str, **kwargs) -> ChatPromptTemplate:
        if system_message is None:
            system_message = (
                "You are an advanced reasoning assistant that judges whether the episodes result in success or failure."
            )

        return ChatPromptTemplate.from_messages(
            [
                ("system", system_message),
                (
                    "human",
                    user_message,
                ),
                MessagesPlaceholder("intermediate_steps"),
                ("human", "Answer: {agent_outcome}"),
                (
                    "human",
                    dedent("""
                     Take your time and comment your decision, but make sure to always output number between 0 and 1 in the end, where 0 would mean 'the episode ended in failure' and 1 would mean 'the episode ended in success'. 
                     Use the following format: [[number]].

                     Your verdict:
                     """),
                ),
            ]
        )

    @classmethod
    def create(
        cls: Type["ReflexionEvaluator"],
        llm: BaseChatModel,
        threshold: float,
        prompt: Optional[ChatPromptTemplate] = None,
        user_message: Optional[str] = None,
        system_message: Optional[str] = None,
        output_parser: Optional[BaseOutputParser[float]] = None,
        parser: Optional[BaseFunctionCallingSingleActionParser | BaseFunctionCallingMultiActionParser] = None,
        parser_name: Optional[str] = None,
    ) -> "ReflexionEvaluator[float]":
        def _preprocess_input(
            inputs: ReflexionEvaluatorInput,
        ) -> Dict:
            # TODO: figure out typing here
            nonlocal parser, parser_name
            if parser is None:
                assert parser_name is not None
                parser = ParserRegistry.get_parser(parser_name)

            preprocessed_inputs = parser.format_inputs(inputs)
            return {
                **preprocessed_inputs["inputs"],
                "agent_outcome": preprocessed_inputs["agent_outcome"].return_values["output"],  # type: ignore[typeddict-item]
                "intermediate_steps": preprocessed_inputs["agent_scratchpad"],
            }

        # TODO: fix typing here
        evaluator: ReflexionEvaluator = cls.create_threshold_evaluator(  # type: ignore[assignment]
            llm=llm,
            threshold=threshold,
            threshold_mode="leq",
            prompt=prompt,
            user_message=user_message,
            system_message=system_message,
            output_parser=output_parser,
        )

        evaluator.add_input_preprocessing(_preprocess_input)

        return evaluator



from typing import Dict, List, Type, Union

from planning_library.function_calling_parsers import (
    BaseFunctionCallingMultiActionParser,
    BaseFunctionCallingSingleActionParser,
)


class ParserRegistry:
    registry: Dict[
        str,
        Union[BaseFunctionCallingSingleActionParser, BaseFunctionCallingMultiActionParser],
    ] = {}

    @classmethod
    def get_parser(
        cls, parser_name
    ) -> Union[BaseFunctionCallingSingleActionParser, BaseFunctionCallingMultiActionParser]:
        try:
            return cls.registry[parser_name]
        except KeyError:
            raise ValueError(f"Unknown parser {parser_name}. Currently available are: {cls.get_available_parsers()}")

    @classmethod
    def get_available_parsers(cls) -> List[str]:
        return list(cls.registry.keys())

    @classmethod
    def register(
        cls,
        other_cls: Union[
            Type[BaseFunctionCallingSingleActionParser],
            Type[BaseFunctionCallingMultiActionParser],
        ],
    ):
        cls.registry[other_cls.name] = other_cls()
        return other_cls



from dataclasses import dataclass, fields
from typing import Dict, List, Optional

from langchain_core.tools import BaseTool


@dataclass
class MetaTools:
    reset: Optional[BaseTool] = None

    @property
    def tools(self) -> List[BaseTool]:
        tools: List[BaseTool] = []

        for field in fields(self):
            field_value = getattr(self, field.name)
            if isinstance(field_value, BaseTool):
                tools.append(field_value)

        return tools

    @property
    def tool_names_map(self) -> Dict[str, str]:
        tool_map: Dict[str, str] = {}

        for field in fields(self):
            field_value = getattr(self, field.name)
            if isinstance(field_value, BaseTool):
                tool_map[field.name] = field_value.name

        return tool_map



from __future__ import annotations

from textwrap import dedent
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

from langchain_core.agents import AgentAction, AgentFinish
from langchain_core.callbacks import AsyncCallbackManager, CallbackManager
from langchain_core.language_models import BaseChatModel
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain_core.runnables import Runnable, RunnableLambda
from langchain_core.tools import BaseTool
from typing_extensions import TypedDict

from planning_library.action_executors import (
    BaseActionExecutor,
    LangchainActionExecutor,
    MetaTools,
)
from planning_library.components import RunnableComponent
from planning_library.components.agent_component import AgentFactory
from planning_library.function_calling_parsers import (
    BaseFunctionCallingMultiActionParser,
    BaseFunctionCallingSingleActionParser,
)
from planning_library.strategies.simple import SimpleStrategy


class ADaPTExecutorInput(TypedDict):
    inputs: Dict[str, Any]


class ADaPTExecutorOutput(TypedDict):
    is_completed: bool
    agent_outcome: AgentFinish
    intermediate_steps: List[Tuple[AgentAction, str]]


class ADaPTExecutor(RunnableComponent[ADaPTExecutorInput, ADaPTExecutorOutput]):
    name = "Executor"

    def __init__(
        self,
        runnable: Runnable[ADaPTExecutorInput, ADaPTExecutorOutput],
        action_executor: BaseActionExecutor,
    ):
        super().__init__(runnable)
        self.add_output_preprocessing(ADaPTExecutor._process_outputs)  # type: ignore[arg-type]
        self._action_executor = action_executor

    @classmethod
    def _create_default_prompt(cls, system_message: Optional[str], user_message: str, **kwargs) -> ChatPromptTemplate:
        if system_message is None:
            system_message = "You are an advanced reasoning agent."

        return ChatPromptTemplate.from_messages(
            [
                ("system", system_message),
                (
                    "human",
                    user_message,
                ),
                (
                    "human",
                    dedent("""
                        When you are sure that you have successfully completed the task,
                         avoid calling tools and make sure to include the words 'task completed' in your output.
                        Do not write 'task completed' if the task has not been completed, for instance,
                         if you think something went wrong and you would like to try again.
                        """),
                ),
                MessagesPlaceholder("agent_scratchpad"),
            ]
        )

    @staticmethod
    def _process_outputs(outputs: Dict[str, Any]) -> ADaPTExecutorOutput:
        outcome = AgentFinish(
            return_values={
                key: value[0]
                for key, value in outputs.items()
                if isinstance(key, list) and key not in ["finish_log", "intermediate_steps"]
            },
            log=outputs["finish_log"][0],
        )

        return {
            "is_completed": "task completed" in outcome.log.lower(),
            "agent_outcome": outcome,
            "intermediate_steps": outputs["intermediate_steps"][0],
        }

    @classmethod
    def create_simple_strategy(
        cls,
        llm: BaseChatModel,
        tools: Optional[Sequence[BaseTool]] = None,
        meta_tools: Optional[MetaTools] = None,
        action_executor: Optional[BaseActionExecutor] = None,
        prompt: Optional[ChatPromptTemplate] = None,
        user_message: Optional[str] = None,
        system_message: Optional[str] = None,
        parser: Optional[
            Union[
                BaseFunctionCallingSingleActionParser,
                BaseFunctionCallingMultiActionParser,
            ]
        ] = None,
        parser_name: Optional[str] = None,
        return_intermediate_steps: bool = False,
        return_finish_log: bool = False,
        max_iterations: int = 15,
        verbose: bool = True,
        **kwargs,
    ) -> "ADaPTExecutor":
        def _preprocess_input(
            inputs: ADaPTExecutorInput,
        ) -> Dict[str, Any]:
            return {**inputs["inputs"]}

        if tools is None:
            assert action_executor is not None, "Either pass tools or action executor."
            tools = action_executor.tools

        prompt = cls._process_prompt(prompt=prompt, user_message=user_message, system_message=system_message)

        agent = AgentFactory.create_agent(llm=llm, tools=tools, prompt=prompt, parser=parser, parser_name=parser_name)

        strategy = SimpleStrategy.create(
            tools=tools,
            action_executor=action_executor
            if action_executor is not None
            else LangchainActionExecutor(tools, meta_tools=meta_tools),
            agent=agent,
            return_intermediate_steps=return_intermediate_steps,
            return_finish_log=return_finish_log,
            max_iterations=max_iterations,
            verbose=verbose,
        )

        runnable = RunnableLambda(_preprocess_input) | strategy

        return cls(runnable=runnable, action_executor=action_executor)  # type: ignore[arg-type]

    def reset(
        self,
        actions: Optional[List[AgentAction]] = None,
        run_manager: Optional[CallbackManager] = None,
        **kwargs,
    ) -> None:
        self._action_executor.reset(actions=actions, run_manager=run_manager, **kwargs)

    async def areset(
        self,
        actions: Optional[List[AgentAction]] = None,
        run_manager: Optional[AsyncCallbackManager] = None,
        **kwargs,
    ) -> None:
        await self._action_executor.areset(actions=actions, run_manager=run_manager, **kwargs)



from typing import Any, Dict, List, Literal, Tuple

from langchain_core.agents import AgentAction, AgentFinish
from langchain_core.pydantic_v1 import BaseModel, Field
from typing_extensions import TypedDict


class ADaPTPlannerInput(TypedDict):
    inputs: Dict[str, Any]
    executor_agent_outcome: AgentFinish
    executor_intermediate_steps: List[Tuple[AgentAction, str]]


class ADaPTPlannerOutput(TypedDict):
    subtasks: List[str]
    aggregation_mode: Literal["and", "or"]


class ADaPTPlan(BaseModel):
    subtasks: List[str] = Field(default_factory=list)
    aggregation_mode: Literal["and", "or"] = Field(default="and")

    def clear(self):
        self.subtasks = []
        self.aggregation_mode = "and"



from typing import Awaitable, Callable, Dict, Generic, Optional, Type

from langchain_core.callbacks import AsyncCallbackManager, CallbackManager
from langchain_core.language_models import BaseChatModel
from langchain_core.output_parsers import BaseOutputParser
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.runnables import Runnable

from planning_library.components.runnable_component import RunnableComponent
from planning_library.primitives.output_parsers import SimpleEvaluateOutputParser

from ..base_component import BaseComponent, InputType, OutputType
from .threshold_judge import GeqThresholdJudge, LeqThresholdJudge


class EvaluatorComponent(Generic[InputType, OutputType], BaseComponent[InputType, bool]):
    def __init__(
        self,
        backbone: BaseComponent[InputType, OutputType],
        judge: BaseComponent[Dict[str, OutputType], bool],
    ):
        self.backbone = backbone
        self.judge = judge

    def add_input_preprocessing(
        self,
        preprocess: Callable[[InputType], Dict],
        apreprocess: Optional[Callable[[InputType], Awaitable[Dict]]] = None,
    ) -> None:
        self.backbone.add_input_preprocessing(preprocess, apreprocess)

    def add_output_preprocessing(
        self,
        preprocess: Callable[[bool], bool],
        apreprocess: Optional[Callable[[bool], Awaitable[bool]]] = None,
    ) -> None:
        self.judge.add_output_preprocessing(preprocess, apreprocess)

    def invoke(self, inputs: InputType, run_manager: Optional[CallbackManager] = None, **kwargs) -> bool:
        if "run_name" not in kwargs and self.name:
            kwargs["run_name"] = self.name

        backbone_output = self.backbone.invoke(inputs, run_manager, **kwargs)
        should_continue = self.judge.invoke({"backbone_output": backbone_output}, run_manager)
        return should_continue

    async def ainvoke(
        self,
        inputs: InputType,
        run_manager: Optional[AsyncCallbackManager] = None,
        **kwargs,
    ) -> bool:
        if "run_name" not in kwargs and self.name:
            kwargs["run_name"] = self.name

        backbone_output = await self.backbone.ainvoke(inputs, run_manager, **kwargs)
        should_continue = await self.judge.ainvoke({"backbone_output": backbone_output}, run_manager)
        return should_continue

    @classmethod
    def create_threshold_evaluator(
        cls: Type["EvaluatorComponent"],
        llm: BaseChatModel,
        threshold: float,
        threshold_mode: str,
        prompt: Optional[ChatPromptTemplate] = None,
        user_message: Optional[str] = None,
        system_message: Optional[str] = None,
        output_parser: Optional[BaseOutputParser[float]] = None,
    ) -> "EvaluatorComponent[InputType, float]":
        prompt = cls._process_prompt(prompt=prompt, user_message=user_message, system_message=system_message)

        if output_parser is None:
            output_parser = SimpleEvaluateOutputParser()

        return cls.create_threshold_evaluator_from_runnable(
            runnable=prompt | llm | output_parser,
            threshold=threshold,
            threshold_mode=threshold_mode,
        )

    @classmethod
    def create_threshold_evaluator_from_runnable(
        cls: Type["EvaluatorComponent"],
        runnable: Runnable[InputType, OutputType],
        threshold: float,
        threshold_mode: str,
    ) -> "EvaluatorComponent[InputType, float]":
        if threshold_mode == "leq":
            judge: BaseComponent[Dict[str, float], bool] = LeqThresholdJudge(threshold=threshold)
        elif threshold_mode == "geq":
            judge = GeqThresholdJudge(threshold=threshold)
        else:
            raise ValueError(f"Unknown `threshold_mode` {threshold_mode} when initializing {cls.__name__}.")

        backbone = RunnableComponent(runnable)
        return cls(backbone=backbone, judge=judge)



from __future__ import annotations

from collections import deque
from typing import (
    AsyncIterator,
    Deque,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
)

from langchain_core.agents import AgentAction, AgentFinish
from langchain_core.callbacks import (
    AsyncCallbackManagerForChainRun,
    CallbackManagerForChainRun,
)

from ...action_executors import BaseActionExecutor, LangchainActionExecutor, MetaTools
from ..base_strategy import BaseCustomStrategy
from .components import (
    ThoughtEvaluator,
    ThoughtEvaluatorConfig,
    ThoughtEvaluatorInput,
    ThoughtGenerator,
    ThoughtGeneratorConfig,
    ThoughtGeneratorInput,
    ThoughtSorter,
    ThoughtSorterConfig,
    ThoughtSorterInput,
)
from .utils import ToTNode


class TreeOfThoughtsDFSStrategy(BaseCustomStrategy):
    """Tree of Thoughts powered by Depth-First Search.

    Based on "Tree of Thoughts: Deliberate Problem Solving with Large Language Models" by Yao et al.

    Also supports DFSDT from "ToolLLM: Facilitating Large Language Models to Master 16000+ Real-world APIs" by Qin et al.
    """

    action_executor: BaseActionExecutor
    thought_generator: ThoughtGenerator
    thought_evaluator: ThoughtEvaluator
    thought_sorter: Optional[ThoughtSorter] = None
    do_sorting: bool = False  # True for DFS (Tree of Thoughts), False for DFSDT (ToolLLM)
    root: Optional[ToTNode] = None
    terminals: List[ToTNode] = []

    @property
    def agent(self):
        return self.thought_generator.agent.agent

    @property
    def input_keys(self) -> List[str]:
        return self.agent.input_keys

    @property
    def output_keys(self) -> List[str]:
        if self.return_intermediate_steps:
            return self.agent.return_values + ["intermediate_steps"]
        else:
            return self.agent.return_values

    @classmethod
    def create(
        cls,
        meta_tools: Optional[MetaTools] = None,
        return_intermediate_steps: bool = False,
        return_finish_log: bool = False,
        max_iterations: int = 15,
        verbose: bool = True,
        action_executor: Optional[BaseActionExecutor] = None,
        generator_config: Optional[ThoughtGeneratorConfig] = None,
        evaluator_config: Optional[ThoughtEvaluatorConfig] = None,
        sorter_config: Optional[ThoughtSorterConfig] = None,
        do_sorting: bool = False,
        **kwargs,
    ) -> "TreeOfThoughtsDFSStrategy":
        """Creates an instance of Tree of Thoughts + DFS strategy.

        Tree of Thoughts + DFS requires an evaluator component. The default setting is as follows:
            * evaluator is a runnable that accepts EvaluatorInput and returns a float in a 0-1 range;
            * evaluator judges whether a new thought should be explored or discarded based on the threshold;
              thought is explored further only when value is greater than the given threshold.

        Args:
            tools: The valid tools the agent can call.
            action_executor: The action executor for the current strategy. If None, the default will be used.
            max_thoughts: Maximum number of new thoughts at each DFS step.
            max_iterations: Maximum number of iterations.
        """
        if generator_config is None:
            raise ValueError("Default thought generator config is currently not supported.")

        if evaluator_config is None:
            raise ValueError("Default thought evaluator config is currently not supported.")

        if do_sorting and sorter_config is None:
            raise ValueError("Default thought sorter config is currently not supported.")

        generator = ThoughtGenerator.create_from_config(generator_config)
        evaluator = ThoughtEvaluator.create_from_config(evaluator_config)
        sorter = ThoughtSorter.create_from_config(sorter_config) if do_sorting else None  # type: ignore[arg-type]

        if action_executor is None:
            action_executor = LangchainActionExecutor(tools=generator_config.tools, meta_tools=meta_tools)

        return cls(
            thought_generator=generator,
            thought_evaluator=evaluator,
            thought_sorter=sorter,
            do_sorting=do_sorting,
            action_executor=action_executor,
            return_intermediate_steps=return_intermediate_steps,
            return_finish_log=return_finish_log,
            max_iterations=max_iterations,
            verbose=verbose,
        )

    def _dfs_step(
        self,
        inputs: Dict[str, str],
        trajectory: List[Tuple[AgentAction, str]],
        run_manager: Optional[CallbackManagerForChainRun] = None,
    ) -> Iterator[List[AgentAction] | AgentAction | AgentFinish]:
        """Performs a single step of DFS algorithm.

        Args:
            inputs: Agent inputs.
            trajectory: Current trajectory – path from the root node to the current node. Essentially, intermediate steps before the current DFS step.
            run_manager: Callback for the current run.

        Returns:
            Iterator over promising thoughts for the current step with three options possible:
              * List[AgentAction] - for multi-action thoughts
              * AgentAction - for single-action thoughts
              * AgentFinish - for finishing thoughts / thoughts without tool calls
        """

        # 1: generate k possible next steps
        thoughts = self.thought_generator.invoke(
            ThoughtGeneratorInput(inputs=inputs, intermediate_steps=trajectory),
            run_manager=run_manager.get_child(tag="generate_thoughts") if run_manager else None,
        )

        # 2: (optional) sort them
        if self.do_sorting:
            assert self.thought_sorter is not None, "Sorting enabled, but thought sorter was not passed."
            thoughts = self.thought_sorter.invoke(
                ThoughtSorterInput(thoughts=thoughts, inputs=inputs, intermediate_steps=trajectory),
                run_manager=run_manager.get_child(tag="sort_thoughts") if run_manager else None,
            )

        for cur_thought in thoughts:
            # 3: evaluate each thought
            cur_thought_should_continue = self.thought_evaluator.invoke(
                ThoughtEvaluatorInput(
                    inputs=inputs,
                    intermediate_steps=trajectory,
                    next_thought=cur_thought,
                ),
                run_manager=run_manager.get_child(tag="evaluate_thought") if run_manager else None,
            )

            # 4: proceed only with thoughts with value above a certain threshold
            if cur_thought_should_continue:
                yield cur_thought

    def _run_strategy(
        self,
        inputs: Dict[str, str],
        run_manager: Optional[CallbackManagerForChainRun] = None,
    ) -> Iterator[Tuple[AgentFinish, List[Tuple[AgentAction, str]]]]:
        """Runs Tree of Thoughts + DFS strategy.

        Args:
            inputs: Agent inputs.
            name_to_tool_map: Mapping from tool names to actual tools, used for calling tools based on agent's output.
            color_mapping: Mapping from tool names to colors, used for logging purposes when calling tools.
            run_manager: Callback for the current run.

        Returns:
            Iterator over tuples (AgentFinish, List[Tuple[AgentAction, str]]):
            essentially, each tuple consists of the final result and of intermediate steps.
            The current implementation iterates over ALL terminal nodes in a tree.
        """
        if not self.root:
            self.root = ToTNode()

        frontier: Deque[ToTNode] = deque([self.root])

        cur_step = 0
        while frontier and cur_step < self.max_iterations:
            cur_node = frontier.pop()

            # TODO: traverses from the tree root to the cur_node on each call. how to optimize?
            trajectory = cur_node.trajectory

            for new_thought in self._dfs_step(
                inputs=inputs,
                trajectory=trajectory,
                run_manager=run_manager,
            ):
                # actually do action(s)
                if isinstance(new_thought, AgentFinish):
                    observation = None
                else:
                    # reset to parent node state
                    self.action_executor.reset(
                        actions=[t[0] for t in trajectory],
                        run_manager=run_manager.get_child() if run_manager else None,
                    )
                    # execute action
                    observation = self.action_executor.execute(
                        actions=new_thought,
                        run_manager=run_manager.get_child() if run_manager else None,
                    )

                new_node = ToTNode(parent=cur_node, thought=new_thought, observation=observation)

                cur_node.children.append(new_node)
                if isinstance(new_thought, AgentFinish):
                    self.terminals.append(new_node)
                else:
                    frontier.appendleft(new_node)

            cur_step += 1

        for node in self.terminals:
            assert isinstance(node.thought, AgentFinish), "Terminal nodes are expected to contain AgentFinish."
            yield node.thought, node.trajectory

    async def _adfs_step(
        self,
        inputs: Dict[str, str],
        trajectory: List[Tuple[AgentAction, str]],
        run_manager: Optional[AsyncCallbackManagerForChainRun] = None,
    ) -> AsyncIterator[List[AgentAction] | AgentAction | AgentFinish]:
        """Performs a single step of DFS algorithm asynchronously.

        Args:
            inputs: Agent inputs.
            trajectory: Current trajectory – path from the root node to the current node. Essentially, intermediate steps before the current DFS step.
            name_to_tool_map: Mapping from tool names to actual tools, used for calling tools based on agent's output.
            color_mapping: Mapping from tool names to colors, used for logging purposes when calling tools.
            run_manager: Callback for the current run.

        Returns:
            Iterator over promising thoughts for the current step with three options possible:
              * List[AgentAction] - for multi-action thoughts
              * AgentAction - for single-action thoughts
              * AgentFinish - for finishing thoughts / thoughts without tool calls
        """

        # 1: generate k possible next steps
        thoughts = await self.thought_generator.ainvoke(
            ThoughtGeneratorInput(inputs=inputs, intermediate_steps=trajectory),
            run_manager=run_manager.get_child(tag="generate_thoughts") if run_manager else None,
        )

        # 2: (optional) sort them
        if self.do_sorting:
            assert self.thought_sorter is not None, "Sorting enabled, but thought sorter was not passed."
            thoughts = await self.thought_sorter.ainvoke(
                ThoughtSorterInput(thoughts=thoughts, inputs=inputs, intermediate_steps=trajectory),
                run_manager=run_manager.get_child(tag="sort_thoughts") if run_manager else None,
            )

        for cur_thought in thoughts:
            # 3: evaluate each thought
            cur_thought_should_continue = await self.thought_evaluator.ainvoke(
                ThoughtEvaluatorInput(
                    inputs=inputs,
                    intermediate_steps=trajectory,
                    next_thought=cur_thought,
                ),
                run_manager=run_manager.get_child(tag="evaluate_thought") if run_manager else None,
            )

            # 4: proceed only with thoughts with value above a certain threshold
            if cur_thought_should_continue:
                yield cur_thought

    async def _arun_strategy(
        self,
        inputs: Dict[str, str],
        run_manager: Optional[AsyncCallbackManagerForChainRun] = None,
    ) -> AsyncIterator[Tuple[AgentFinish, List[Tuple[AgentAction, str]]]]:
        """Runs Tree of Thoughts + DFS strategy asynchronously.

        Args:
            inputs: Agent inputs.
            name_to_tool_map: Mapping from tool names to actual tools, used for calling tools based on agent's output.
            color_mapping: Mapping from tool names to colors, used for logging purposes when calling tools.
            run_manager: Callback for the current run.

        Returns:
            Iterator over tuples (AgentFinish, List[Tuple[AgentAction, str]]):
            essentially, each tuple consists of the final result and of intermediate steps.
            The current implementation iterates over ALL terminal nodes in a tree.
        """
        if not self.root:
            self.root = ToTNode()

        frontier: Deque[ToTNode] = deque([self.root])

        cur_step = 0
        while frontier and cur_step < self.max_iterations:
            cur_node = frontier.pop()

            # TODO: traverses from the tree root to the cur_node on each call. how to optimize?
            trajectory = cur_node.trajectory

            async for new_thought in self._adfs_step(
                inputs=inputs,
                trajectory=trajectory,
                run_manager=run_manager,
            ):
                # actually do action(s)
                if isinstance(new_thought, AgentFinish):
                    observation = None
                else:
                    # reset to parent node state
                    await self.action_executor.areset(
                        actions=[t[0] for t in trajectory],
                        run_manager=run_manager.get_child() if run_manager else None,
                    )
                    # execute action
                    observation = await self.action_executor.aexecute(
                        actions=new_thought,
                        run_manager=run_manager.get_child() if run_manager else None,
                    )

                new_node = ToTNode(parent=cur_node, thought=new_thought, observation=observation)
                cur_node.children.append(new_node)
                if isinstance(new_thought, AgentFinish):
                    self.terminals.append(new_node)
                else:
                    frontier.appendleft(new_node)

            cur_step += 1

        for node in self.terminals:
            assert isinstance(node.thought, AgentFinish), "Terminal nodes are expected to contain AgentFinish."
            yield node.thought, node.trajectory



from __future__ import annotations

from textwrap import dedent
from typing import Any, Dict, List, Optional, Sequence, Tuple

from langchain_core.agents import AgentAction
from langchain_core.messages import BaseMessage
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from typing_extensions import TypedDict

from planning_library.components import AgentComponent


class ReflexionActorInput(TypedDict):
    inputs: Dict[str, Any]
    intermediate_steps: List[Tuple[AgentAction, str]]
    self_reflections: Sequence[BaseMessage]


class ReflexionActor(AgentComponent[ReflexionActorInput]):
    """Actor component for Reflexion strategy.

    This component is powered by the agent and expects ReflexionActorInput to be passed as input.

    In addition to default agent initialization, this component also provides a default prompt template, which can
    be initialized with only a single user message.
    """

    name = "Actor"

    required_prompt_input_vars = set(ReflexionActorInput.__annotations__) - {
        "inputs",
        "intermediate_steps",
    } | {"agent_scratchpad"}

    @classmethod
    def _create_default_prompt(cls, system_message: Optional[str], user_message: str, **kwargs) -> ChatPromptTemplate:
        if system_message is None:
            system_message = "You are an advanced reasoning agent that can improve based on self-reflection."

        return ChatPromptTemplate.from_messages(
            [
                ("system", system_message),
                (
                    "human",
                    user_message,
                ),
                (
                    "human",
                    dedent("""
                        This might be not your first attempt. 
                        In this case, you will find self-reflective thoughts below. 
                        Make sure to pay extra attention to them, as they aim to identify and mitigate the exact shortcomings that led to failure in previous trials. 
                        """),
                ),
                MessagesPlaceholder("self_reflections"),
                (
                    "human",
                    "Good luck!",
                ),
                MessagesPlaceholder("agent_scratchpad"),
            ]
        )



from __future__ import annotations

from dataclasses import dataclass
from textwrap import dedent
from typing import Any, Dict, Generic, List, Optional, Tuple, Type, Union

from langchain_core.agents import AgentAction, AgentFinish
from langchain_core.language_models import BaseChatModel
from langchain_core.output_parsers import BaseOutputParser
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain_core.runnables import Runnable
from typing_extensions import TypedDict

from planning_library.components.base_component import OutputType
from planning_library.components.evaluation import EvaluatorComponent
from planning_library.function_calling_parsers import (
    BaseFunctionCallingMultiActionParser,
    BaseFunctionCallingSingleActionParser,
    ParserRegistry,
)
from planning_library.utils import (
    format_thought,
)


@dataclass
class ThoughtEvaluatorConfig:
    value_threshold: float

    runnable: Optional[Runnable] = None

    prompt: Optional[ChatPromptTemplate] = None
    user_message: Optional[str] = None
    system_message: Optional[str] = None

    llm: Optional[BaseChatModel] = None

    parser: Optional[
        Union[
            BaseFunctionCallingSingleActionParser,
            BaseFunctionCallingMultiActionParser,
        ]
    ] = None
    parser_name: Optional[str] = None

    output_parser: Optional[BaseOutputParser[float]] = None


class ThoughtEvaluatorInput(TypedDict):
    inputs: Dict[str, Any]
    intermediate_steps: List[Tuple[AgentAction, str]]
    next_thought: List[AgentAction] | AgentAction | AgentFinish


class ThoughtEvaluator(Generic[OutputType], EvaluatorComponent[ThoughtEvaluatorInput, OutputType]):
    name = "Evaluate Thoughts"

    required_prompt_input_vars = set(ThoughtEvaluatorInput.__annotations__) - {"inputs"}

    @classmethod
    def _create_default_prompt(cls, system_message: Optional[str], user_message: str, **kwargs) -> ChatPromptTemplate:
        if system_message is None:
            system_message = (
                "You are an advanced reasoning assistant that judges the plausability of "
                "steps suggested for solving complex reasoning tasks."
            )

        return ChatPromptTemplate.from_messages(
            [
                ("system", system_message),
                (
                    "human",
                    user_message,
                ),
                MessagesPlaceholder("intermediate_steps"),
                ("human", "Here is the proposed next step:" ""),
                MessagesPlaceholder("next_thought"),
                (
                    "human",
                    dedent("""
                     Your goal is to judge whether this proposal should be followed or discarded, 
                     how likely it is to lead to the success.
                     
                     Take your time and comment your decision, 
                     but make sure to always output number between 0 and 1 in the end, 
                     where 0 would mean 'the proposed action is incorrect and/or very unlikely to lead to success' 
                     and 1 would mean 'the proposed action is correct and very likely to lead to success'. 
                     
                     ALWAYS use the following format and add the number in the end of your answer: [[number]].

                     Your verdict:
                     """),
                ),
            ]
        )

    @classmethod
    def create(
        cls: Type["ThoughtEvaluator"],
        llm: BaseChatModel,
        threshold: float,
        prompt: Optional[ChatPromptTemplate] = None,
        user_message: Optional[str] = None,
        system_message: Optional[str] = None,
        output_parser: Optional[BaseOutputParser[float]] = None,
        parser: Optional[BaseFunctionCallingSingleActionParser | BaseFunctionCallingMultiActionParser] = None,
        parser_name: Optional[str] = None,
    ) -> "ThoughtEvaluator[float]":
        def _preprocess_input(
            inputs: ThoughtEvaluatorInput,
        ) -> Dict:
            # TODO: figure out typing here
            nonlocal parser, parser_name
            if parser is None:
                assert parser_name is not None
                parser = ParserRegistry.get_parser(parser_name)

            intermediate_steps = parser.format_inputs(inputs)["agent_scratchpad"]

            return {
                **inputs["inputs"],
                "next_thought": format_thought(inputs["next_thought"]),
                "intermediate_steps": intermediate_steps,
            }

        # TODO: fix typing here
        evaluator: ThoughtEvaluator = cls.create_threshold_evaluator(  # type: ignore[assignment]
            llm=llm,
            threshold=threshold,
            threshold_mode="geq",
            prompt=prompt,
            user_message=user_message,
            system_message=system_message,
            output_parser=output_parser,
        )

        evaluator.add_input_preprocessing(_preprocess_input)

        return evaluator

    @classmethod
    def create_from_config(cls, config: ThoughtEvaluatorConfig) -> ThoughtEvaluator:
        if config.runnable is not None:
            evaluator: ThoughtEvaluator = cls.create_threshold_evaluator_from_runnable(  # type: ignore[assignment]
                runnable=config.runnable,
                threshold=config.value_threshold,
                threshold_mode="geq",
            )
            return evaluator

        if config.llm is None:
            raise ValueError("`llm` must be provided when `runnable` is None.")

        return cls.create(
            llm=config.llm,
            prompt=config.prompt,
            user_message=config.user_message,
            system_message=config.system_message,
            threshold=config.value_threshold,
            output_parser=config.output_parser,
            parser=config.parser,
            parser_name=config.parser_name,
        )



from typing import List, Optional, Tuple, Union

from langchain_core.agents import AgentAction, AgentFinish, AgentStep


class ToTNode:
    def __init__(
        self,
        parent: Optional["ToTNode"] = None,
        thought: Optional[Union[List[AgentAction], AgentAction, AgentFinish]] = None,
        observation: Optional[Union[List[AgentStep], AgentStep]] = None,
    ):
        self.parent = parent
        self.children: List["ToTNode"] = []
        self.thought = thought
        self.observation = observation

    @property
    def trajectory(self) -> List[Tuple[AgentAction, str]]:
        """Returns the (action, observation) tuples on the path from the root to the current node.

        Note:
            * The nodes are arranged in order from the root to the current node.
        """
        node: Optional[ToTNode] = self
        trajectory_actions: List[Tuple[AgentAction, str]] = []
        while node is not None:
            if isinstance(node.thought, list):
                assert isinstance(node.observation, list) and len(node.thought) == len(node.observation)
                trajectory_actions.extend(
                    (observation.action, observation.observation) for observation in node.observation
                )
            elif isinstance(node.thought, AgentAction):
                assert isinstance(node.observation, AgentStep)
                trajectory_actions.append((node.observation.action, node.observation.observation))
            elif isinstance(node.thought, AgentFinish) and node is not self:
                raise ValueError("AgentFinish detected as non-terminal node.")

            node = node.parent
        return trajectory_actions[::-1]



from __future__ import annotations

from typing import List

from langchain_core.agents import AgentAction, AgentFinish
from langchain_core.messages import AIMessage, BaseMessage


def format_thought(
    thought: List[AgentAction] | AgentAction | AgentFinish,
) -> List[BaseMessage]:
    if isinstance(thought, list):
        messages = []
        for action in thought:
            messages.extend(format_thought(action))
        return messages
    elif isinstance(thought, AgentAction):
        return [AIMessage(content=f"Call tool `{thought.tool}` with arguments `{thought.tool_input}`")]
    elif isinstance(thought, AgentFinish):
        return [AIMessage(content=f"Finish execution with return values `{thought.return_values}`")]

    raise ValueError(f"Unexpected type for `thought`: {type(thought)}")


def format_thoughts(
    thoughts: List[List[AgentAction] | AgentAction | AgentFinish],
) -> List[BaseMessage]:
    messages = []
    for thought in thoughts:
        messages.extend(format_thought(thought))
    return messages



from .thought_evaluator import (
    ThoughtEvaluator,
    ThoughtEvaluatorConfig,
    ThoughtEvaluatorInput,
)
from .thought_generator import (
    ThoughtGenerator,
    ThoughtGeneratorConfig,
    ThoughtGeneratorInput,
)
from .thought_sorter import ThoughtSorter, ThoughtSorterConfig, ThoughtSorterInput

__all__ = [
    "ThoughtGeneratorInput",
    "ThoughtGenerator",
    "ThoughtSorterInput",
    "ThoughtSorter",
    "ThoughtEvaluatorInput",
    "ThoughtEvaluator",
    "ThoughtEvaluatorConfig",
    "ThoughtSorterConfig",
    "ThoughtGeneratorConfig",
]
